"""
Data types to describe events.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.date import HasDate
from betty.ancestry.description import HasDescription
from betty.ancestry.has_citations import HasCitations
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.place import Place
from betty.ancestry.presence import Presence
from betty.event_type import EventTypeDefinition
from betty.event_type.event_types import Unknown as UnknownEventType
from betty.json.linked_data import JsonLdObject, dump_context
from betty.json.schema import String
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.markup import AllEnumeration
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema
from betty.model import EntityDefinition
from betty.model.association import (
    BidirectionalToManySingleType,
    BidirectionalToZeroOrOne,
    ToManyAssociates,
    ToZeroOrOneAssociate,
)
from betty.plugin.schema import PluginIdSchema
from betty.privacy import HasPrivacy, Privacy
from betty.property import Optional
from betty.role.roles import Subject

if TYPE_CHECKING:
    from collections.abc import Mapping

    from betty.ancestry.citation import Citation
    from betty.ancestry.file_reference import FileReference
    from betty.ancestry.note import Note
    from betty.date import ResolvableDate
    from betty.event_type import EventType
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "event",
    label=_("Event"),
    label_plural=_("Events"),
    label_countable=ngettext("{count} event", "{count} events"),
)
class Event(
    HasDate,
    HasFileReferences,
    HasCitations,
    HasNotes,
    HasDescription,
    HasPrivacy,
    HasLinks,
):
    """
    .. plugin:: entity:event.
    """

    name = Optional(LocalizableProperty(label=_("Name")))
    """
    The event's name, if it has any.
    """

    place = BidirectionalToZeroOrOne["Event", Place](
        "betty.ancestry.place:Place",
        "events",
        label=_("Place"),
        description=_("The location of the event"),
    )
    """
    The place the event happened.
    """
    presences = BidirectionalToManySingleType["Event", Presence](
        "betty.ancestry.presence:Presence",
        "event",
        label=_("Presences"),
        description=_("People's presences at this event"),
        linked_data_embedded=True,
    )
    """
    People's presences at this event.
    """

    def __init__(
        self,
        *,
        id: str | None = None,  # noqa: A002
        event_type: EventType | None = None,
        date: ResolvableDate | None = None,
        file_references: ToManyAssociates[FileReference] | None = None,
        citations: ToManyAssociates[Citation] | None = None,
        notes: ToManyAssociates[Note] | None = None,
        privacy: Privacy | None = None,
        place: ToZeroOrOneAssociate[Place] = None,
        description: ResolvableLocalizable | None = None,
        presences: ToManyAssociates[Presence] | None = None,
        name: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            id,
            date=date,
            file_references=file_references,
            citations=citations,
            notes=notes,
            privacy=privacy,
            description=description,
        )
        self._event_type = event_type or UnknownEventType()
        if place is not None:
            self.place = place
        if presences is not None:
            self.presences = presences
        self.name = name

    @override
    def dated_linked_data_contexts(self) -> tuple[str | None, str | None, str | None]:
        return (
            "https://schema.org/startDate",
            "https://schema.org/startDate",
            "https://schema.org/endDate",
        )

    @override
    @property
    def label(self) -> Localizable:
        if self.name:
            return self.name

        format_kwargs: Mapping[str, ResolvableLocalizable] = {
            "event_type": self._event_type.plugin().label,
        }
        subjects = [
            presence.person
            for presence in self.presences
            if presence.public
            and isinstance(presence.role, Subject)
            and presence.person.public
        ]
        if subjects:
            format_kwargs["subjects"] = AllEnumeration(
                *(person.label for person in subjects)
            )

        if subjects:
            return _("{event_type} of {subjects}").format(**format_kwargs)
        return _("{event_type}").format(**format_kwargs)

    @property
    def event_type(self) -> EventType:
        """
        The type of event.
        """
        return self._event_type

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        dump_context(portable, place="https://schema.org/location")
        dump_context(portable, presences="https://schema.org/performer")
        portable["@type"] = "https://schema.org/Event"
        portable["type"] = self.event_type.plugin().id
        portable["eventAttendanceMode"] = (
            "https://schema.org/OfflineEventAttendanceMode"
        )
        portable["eventStatus"] = "https://schema.org/EventScheduled"
        if self.name is not None:
            portable["name"] = dump_linked_data(
                self.name, localizers=await project.public_localizers
            )
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "name",
            StaticTranslationsSchema(),
            False,
        )
        schema.add_property(
            "type",
            PluginIdSchema(
                EventTypeDefinition.type(),
                await project.plugins.plugins(EventTypeDefinition),
            ),
        )
        schema.add_property("eventStatus", String(title="Event status"))
        schema.add_property(
            "eventAttendanceMode", String(title="Event attendance mode")
        )
        return schema



"""
Gender configuration.
"""

from __future__ import annotations

from typing import final, override

from betty.data.aggregate.record.object import ObjectDefinition
from betty.gender import Gender, GenderDefinition
from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.gettext import _
from betty.locale.localizable.static import CountableStaticTranslations
from betty.plugin.data import CountableHumanFacingPluginDefinitionConfiguration
from betty.sample import Sample


@final
@ObjectDefinition(
    label=_("Gender configuration"),
    samples=[
        lambda: Sample(
            GenderDefinitionConfiguration(
                id="genderqueer",
                label="Genderqueer",
                label_plural="Genderqueers",
                label_countable=CountableStaticTranslations(
                    {
                        DEFAULT_LOCALE: {
                            "one": "{count} genderqueer",
                            "other": "{count} genderqueers",
                        }
                    }
                ),
            ),
            label="Default",
        )
    ],
)
class GenderDefinitionConfiguration(
    CountableHumanFacingPluginDefinitionConfiguration[GenderDefinition]
):
    """
    Configure a :py:class:`betty.gender.GenderDefinition`.

    .. data:: betty.project.data:GenderDefinitionConfiguration
    """

    @override
    def new_plugin(self) -> GenderDefinition:
        @GenderDefinition(
            self.id,
            label=self.label,
            label_plural=self.label_plural,
            label_countable=self.label_countable,
            description=self.description,
        )
        class _ProjectConfigurationGender(Gender):
            pass

        return _ProjectConfigurationGender.plugin()



"""
Access discovered plugins.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from betty.plugin import PluginDefinition

if TYPE_CHECKING:
    import builtins
    from collections.abc import Iterator

    from betty.machine_name import ResolvableMachineName


class PluginRepository[PluginDefinitionT: PluginDefinition = PluginDefinition](ABC):
    """
    Access discovered plugins.
    """

    def __init__(self, plugin_type: builtins.type[PluginDefinitionT]):
        self._type = plugin_type

    @property
    def type(self) -> builtins.type[PluginDefinitionT]:
        """
        The plugin type contained by this repository.
        """
        return self._type

    @abstractmethod
    def get(self, plugin_id: ResolvableMachineName, /) -> PluginDefinitionT:
        """
        Get a single plugin by its ID.

        :raises PluginUnavailable: if no plugin can be found for the given ID.
        """

    def __len__(self) -> int:
        return len(list(self.__iter__()))

    @abstractmethod
    def __iter__(self) -> Iterator[PluginDefinitionT]:
        pass

    def __getitem__(self, plugin_id: ResolvableMachineName) -> PluginDefinitionT:
        return self.get(plugin_id)



"""
Localize dates.
"""

from __future__ import annotations

import calendar
import datetime
import operator
from contextlib import suppress
from functools import total_ordering
from typing import TYPE_CHECKING, Any, override

from babel import dates

from betty.locale.localizable import Localizable
from betty.locale.localizable.gettext import _

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from types import NotImplementedType

    from ty_extensions import Intersection

    from betty.locale import HasLocale
    from betty.locale.localize import Localizer


class IncompleteDateError(ValueError):
    """
    Raised when a date-like was unexpectedly incomplete.
    """


_LOCALIZE_DATE_PART_FORMATS: Mapping[tuple[bool, bool, bool], Localizable] = {
    (True, True, True): _("MMMM d, y"),
    (True, True, False): _("MMMM, y"),
    (True, False, False): _("y"),
    (False, True, True): _("MMMM d"),
    (False, True, False): _("MMMM"),
}


def _localize_date_parts(localizer: Localizer, date: Date | None, /) -> str:
    if date is None:
        raise IncompleteDateError("This date is None.")
    try:
        date_parts_format = _LOCALIZE_DATE_PART_FORMATS[
            (date.year is not None, date.month is not None, date.day is not None)
        ].localize(localizer)
    except KeyError:
        raise IncompleteDateError(
            "This date does not have enough parts to be rendered."
        ) from None
    parts = (1 if x is None else x for x in date.parts)
    return dates.format_date(datetime.date(*parts), date_parts_format, localizer.locale)


class Date(Localizable):
    """
    A (Gregorian) date.
    """

    _LOCALIZE_FORMATS: Mapping[tuple[bool | None], Localizable] = {
        (True,): _("around {date}"),
        (False,): _("{date}"),
    }

    year: int | None
    month: int | None
    day: int | None
    fuzzy: bool

    def __init__(
        self,
        year: int | None = None,
        month: int | None = None,
        day: int | None = None,
        *,
        fuzzy: bool = False,
    ):
        self.year = year
        self.month = month
        self.day = day
        self.fuzzy = fuzzy

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        try:
            return (
                self._LOCALIZE_FORMATS[(self.fuzzy,)]
                .format(
                    date=_localize_date_parts(localizer, self),
                )
                .localize(localizer)
            )
        except IncompleteDateError:
            return _("unknown date").localize(localizer)

    @property
    def comparable(self) -> bool:
        """
        If this date is comparable to other date-likes.
        """
        return self.year is not None

    @property
    def complete(self) -> bool:
        """
        Whether this date is complete.
        """
        return self.year is not None and self.month is not None and self.day is not None

    @property
    def parts(self) -> tuple[int | None, int | None, int | None]:
        """
        The date parts: a 3-tuple of the year, month, and day.
        """
        return self.year, self.month, self.day

    def to_range(self) -> DateRange:
        """
        Convert this date to a date range.
        """
        if not self.comparable:
            raise ValueError(
                f"Cannot convert non-comparable date {repr(self)} to a date range."
            )
        assert self.year is not None
        if self.month is None:
            month_start = 1
            month_end = 12
        else:
            month_start = month_end = self.month
        if self.day is None:
            day_start = 1
            day_end = calendar.monthrange(self.year, month_end)[1]
        else:
            day_start = day_end = self.day
        return DateRange(
            Date(self.year, month_start, day_start), Date(self.year, month_end, day_end)
        )

    def _compare(
        self, other: Any, comparator: Callable[[Any, Any], bool], /
    ) -> bool | NotImplementedType:
        if not isinstance(other, Date):
            return NotImplemented
        selfish = self
        if not selfish.comparable or not other.comparable:
            return NotImplemented
        if selfish.complete and other.complete:
            return comparator(selfish.parts, other.parts)
        if not other.complete:
            other = other.to_range()
        if not selfish.complete:
            selfish = selfish.to_range()
        return comparator(selfish, other)

    def __contains__(self, other: ResolvableDate) -> bool:
        if isinstance(other, Date):
            return self == other
        return self in other

    def __lt__(self, other: Any) -> bool:
        return self._compare(other, operator.lt)

    def __le__(self, other: Any) -> bool:
        return self._compare(other, operator.le)

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Date):
            return NotImplemented
        return self.parts == other.parts

    def __ge__(self, other: Any) -> bool:
        return self._compare(other, operator.ge)

    def __gt__(self, other: Any) -> bool:
        return self._compare(other, operator.gt)


def _dump_date_iso8601(date: Date, /) -> str | None:
    if not date.complete:
        return None
    assert date.year
    assert date.month
    assert date.day
    return f"{date.year:04d}-{date.month:02d}-{date.day:02d}"


@total_ordering
class DateRange(Localizable):
    """
    A date range can describe a period of time between, before, after, or around start and/or end dates.
    """

    _LOCALIZE_FORMATS: Mapping[
        tuple[bool | None, bool | None, bool | None, bool | None], Localizable
    ] = {
        (False, False, False, False): _("from {start_date} until {end_date}"),
        (False, False, False, True): _(
            "from {start_date} until sometime before {end_date}"
        ),
        (False, False, True, False): _("from {start_date} until around {end_date}"),
        (False, False, True, True): _(
            "from {start_date} until sometime before around {end_date}"
        ),
        (False, True, False, False): _(
            "from sometime after {start_date} until {end_date}"
        ),
        (False, True, False, True): _("sometime between {start_date} and {end_date}"),
        (False, True, True, False): _(
            "from sometime after {start_date} until around {end_date}"
        ),
        (False, True, True, True): _(
            "sometime between {start_date} and around {end_date}"
        ),
        (True, False, False, False): _("from around {start_date} until {end_date}"),
        (True, False, False, True): _(
            "from around {start_date} until sometime before {end_date}"
        ),
        (True, False, True, False): _(
            "from around {start_date} until around {end_date}"
        ),
        (True, False, True, True): _(
            "from around {start_date} until sometime before around {end_date}"
        ),
        (True, True, False, False): _(
            "from sometime after around {start_date} until {end_date}"
        ),
        (True, True, False, True): _(
            "sometime between around {start_date} and {end_date}"
        ),
        (True, True, True, False): _(
            "from sometime after around {start_date} until around {end_date}"
        ),
        (True, True, True, True): _(
            "sometime between around {start_date} and around {end_date}"
        ),
        (False, False, None, None): _("from {start_date}"),
        (False, True, None, None): _("sometime after {start_date}"),
        (True, False, None, None): _("from around {start_date}"),
        (True, True, None, None): _("sometime after around {start_date}"),
        (None, None, False, False): _("until {end_date}"),
        (None, None, False, True): _("sometime before {end_date}"),
        (None, None, True, False): _("until around {end_date}"),
        (None, None, True, True): _("sometime before around {end_date}"),
    }

    start: Date | None
    start_is_boundary: bool
    end: Date | None
    end_is_boundary: bool

    def __init__(
        self,
        start: Date | None = None,
        end: Date | None = None,
        *,
        start_is_boundary: bool = False,
        end_is_boundary: bool = False,
    ):
        self.start = start
        self.start_is_boundary = start_is_boundary
        self.end = end
        self.end_is_boundary = end_is_boundary

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        formatter_configuration: tuple[
            bool | None, bool | None, bool | None, bool | None
        ] = (None, None, None, None)
        formatter_arguments = {}

        with suppress(IncompleteDateError):
            formatter_arguments["start_date"] = _localize_date_parts(
                localizer, self.start
            )
            formatter_configuration = (
                None if self.start is None else self.start.fuzzy,
                self.start_is_boundary,
                formatter_configuration[2],
                formatter_configuration[3],
            )

        with suppress(IncompleteDateError):
            formatter_arguments["end_date"] = _localize_date_parts(localizer, self.end)
            formatter_configuration = (
                formatter_configuration[0],
                formatter_configuration[1],
                None if self.end is None else self.end.fuzzy,
                self.end_is_boundary,
            )

        if not formatter_arguments:
            raise IncompleteDateError(
                "This date range does not have enough parts to be rendered."
            )

        return (
            self._LOCALIZE_FORMATS[formatter_configuration]
            .format(**formatter_arguments)
            .localize(localizer)
        )

    @property
    def comparable(self) -> bool:
        """
        If this date is comparable to other date-likes.
        """
        return (
            self.start is not None
            and self.start.comparable
            or self.end is not None
            and self.end.comparable
        )

    def __contains__(self, other: ResolvableDate) -> bool:
        if not self.comparable:
            return False

        if isinstance(other, Date):
            others = [other]
        else:
            if not other.comparable:
                return False
            others = []
            if other.start is not None and other.start.comparable:
                others.append(other.start)
            if other.end is not None and other.end.comparable:
                others.append(other.end)

        if self.start is not None and self.end is not None:
            if isinstance(other, DateRange) and (
                other.start is None or other.end is None
            ):
                if other.start is None:
                    return self.start <= other.end or self.end <= other.end
                if other.end is None:
                    return self.start >= other.start or self.end >= other.start
            for another in others:
                if self.start <= another <= self.end:
                    return True
            if isinstance(other, DateRange):
                for selfdate in [self.start, self.end]:
                    if other.start <= selfdate <= other.end:
                        return True

        elif self.start is not None:
            # Two date ranges with start dates only always overlap.
            if isinstance(other, DateRange) and other.end is None:
                return True

            for other in others:
                if self.start <= other:
                    return True
        elif self.end is not None:
            # Two date ranges with end dates only always overlap.
            if isinstance(other, DateRange) and other.start is None:
                return True

            for other in others:
                if other <= self.end:
                    return True
        return False

    def _get_comparable_date(self, date: Date | None, /) -> Date | None:
        if date and date.comparable:
            return date
        return None

    _LT_DATE_RANGE_COMPARATORS = {
        (
            True,
            True,
            True,
            True,
        ): lambda self_start, self_end, other_start, other_end: (
            self_start < other_start
        ),
        (
            True,
            True,
            True,
            False,
        ): lambda self_start, self_end, other_start, other_end: (
            self_start <= other_start
        ),
        (
            True,
            True,
            False,
            True,
        ): lambda self_start, self_end, other_start, other_end: (
            self_start < other_end or self_end <= other_end
        ),
        (
            True,
            True,
            False,
            False,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (
            True,
            False,
            True,
            True,
        ): lambda self_start, self_end, other_start, other_end: (
            self_start < other_start
        ),
        (
            True,
            False,
            True,
            False,
        ): lambda self_start, self_end, other_start, other_end: (
            self_start < other_start
        ),
        (
            True,
            False,
            False,
            True,
        ): lambda self_start, self_end, other_start, other_end: self_start < other_end,
        (
            True,
            False,
            False,
            False,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (
            False,
            True,
            True,
            True,
        ): lambda self_start, self_end, other_start, other_end: self_end <= other_start,
        (
            False,
            True,
            True,
            False,
        ): lambda self_start, self_end, other_start, other_end: self_end <= other_start,
        (
            False,
            True,
            False,
            True,
        ): lambda self_start, self_end, other_start, other_end: self_end < other_end,
        (
            False,
            True,
            False,
            False,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (
            False,
            False,
            True,
            True,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (
            False,
            False,
            True,
            False,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (
            False,
            False,
            False,
            True,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (
            False,
            False,
            False,
            False,
        ): lambda self_start, self_end, other_start, other_end: NotImplemented,
    }

    _LT_DATE_COMPARATORS = {
        (True, True): lambda self_start, self_end, other: self_start < other,
        (True, False): lambda self_start, self_end, other: self_start < other,
        (False, True): lambda self_start, self_end, other: self_end <= other,
        (False, False): lambda self_start, self_end, other: NotImplemented,
    }

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Date | DateRange):
            return NotImplemented

        self_start = self._get_comparable_date(self.start)
        self_end = self._get_comparable_date(self.end)
        signature = (
            self_start is not None,
            self_end is not None,
        )
        if isinstance(other, DateRange):
            other_start = self._get_comparable_date(other.start)
            other_end = self._get_comparable_date(other.end)
            return self._LT_DATE_RANGE_COMPARATORS[
                (
                    *signature,
                    other_start is not None,
                    other_end is not None,
                )
            ](self_start, self_end, other_start, other_end)
        return self._LT_DATE_COMPARATORS[signature](self_start, self_end, other)

    @override
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Date):
            return False

        if not isinstance(other, DateRange):
            return NotImplemented
        return (self.start, self.end, self.start_is_boundary, self.end_is_boundary) == (
            other.start,
            other.end,
            other.start_is_boundary,
            other.end_is_boundary,
        )


type ResolvableDate = Date | DateRange



"""
Console user sessions.
"""

import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import TextIO, cast, final, overload, override

from rich.console import Console
from rich.progress import BarColumn, TaskProgressColumn, TextColumn, TimeElapsedColumn
from rich.progress import Progress as _RichProgress
from rich.prompt import Confirm, Prompt

from betty.assertion import Assertion
from betty.life_cycle.manage import ManagedLifeCycle
from betty.locale.localizable import ResolvableLocalizable
from betty.locale.localize import resolve_localized
from betty.progress import Progress
from betty.progress.no_op import NoOpProgress
from betty.rich import Theme
from betty.rich.progress import RichProgress
from betty.typing import Void, VoidType
from betty.user import User, Verbosity
from betty.user.logging import UserHandler


@final
class RichUser(ManagedLifeCycle, User):
    """
    A Rich user session.
    """

    def __init__(self):
        super().__init__()
        self.life_cycle.on_bootstrap(self._propagate_verbosity)
        self.life_cycle.on_shutdown(self._shutdown_logging_handler)
        self._console = Console(theme=Theme())
        self._verbosity = Verbosity.DEFAULT
        self._logging_handler: UserHandler | None = None
        self._logger = logging.getLogger()
        self._log_formatter = logging.Formatter()

    @property
    def console(self) -> Console:
        """
        The Rich console.
        """
        return self._console

    @override
    @property
    def verbosity(self) -> Verbosity:
        return self._verbosity

    @override
    async def set_verbosity(self, verbosity: Verbosity, /) -> None:
        if verbosity is self._verbosity:
            return
        self._verbosity = verbosity
        if self.bootstrapped:
            await self._propagate_verbosity()

    async def _shutdown_logging_handler(self, *, wait: bool = True) -> None:
        if self._logging_handler is not None:
            await self._logging_handler.shutdown(wait=wait)

    async def _propagate_verbosity(self) -> None:
        if self.verbosity >= Verbosity.MOST_VERBOSE:
            if self._logging_handler is not None:
                return

            self._logging_handler = UserHandler(self)
            self._logger.addHandler(self._logging_handler)
            await self._logging_handler.bootstrap()
            level = logging.NOTSET
        else:
            if self._logging_handler is None:
                return
            self._logger.removeHandler(self._logging_handler)
            await self._logging_handler.shutdown()
            level = 999999999
        self._logger.setLevel(level)

    @override
    async def message_exception(self) -> None:
        self._message_error(self.localizer._("An unexpected error occurred:"))
        self._console.print_exception(show_locals=self.verbosity >= Verbosity.VERBOSE)

    @override
    async def message_error(self, message: ResolvableLocalizable, /) -> None:
        self._message_error(resolve_localized(message, localizer=self.localizer))

    def _message_error(self, message: str) -> None:
        self.assert_alive()
        self._console.print(f"[red]{message}[/]")

    @override
    async def message_warning(self, message: ResolvableLocalizable, /) -> None:
        self.assert_alive()
        if self._verbosity < Verbosity.DEFAULT:
            return
        self._console.print(
            f"[yellow]{resolve_localized(message, localizer=self.localizer)}[/]"
        )

    @override
    async def message_information(self, message: ResolvableLocalizable, /) -> None:
        self.assert_alive()
        if self._verbosity < Verbosity.DEFAULT:
            return
        self._console.print(
            f"[green]{resolve_localized(message, localizer=self.localizer)}[/]"
        )

    @override
    async def message_information_details(
        self, message: ResolvableLocalizable, /
    ) -> None:
        self.assert_alive()
        if self._verbosity < Verbosity.VERBOSE:
            return
        self._console.print(
            f"[green]{resolve_localized(message, localizer=self.localizer)}[/]"
        )

    @override
    async def message_debug(self, message: ResolvableLocalizable, /) -> None:
        self.assert_alive()
        if self._verbosity < Verbosity.MORE_VERBOSE:
            return
        self._console.print(
            f"[white]{resolve_localized(message, localizer=self.localizer)}[/]"
        )

    @override
    async def message_log(self, message: logging.LogRecord, /) -> None:
        self.assert_bootstrapped()
        if self._verbosity < Verbosity.MOST_VERBOSE:
            return
        self._console.print(f"[blue]{self._log_formatter.format(message)}[/]")

    @override
    @asynccontextmanager
    async def message_progress(
        self, message: ResolvableLocalizable, /
    ) -> AsyncIterator[Progress]:
        self.assert_alive()
        if self.verbosity < Verbosity.DEFAULT:
            yield NoOpProgress()
        else:
            with _RichProgress(
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TaskProgressColumn(),
                TimeElapsedColumn(),
                console=self._console,
            ) as rich_progress:
                yield RichProgress(
                    rich_progress, resolve_localized(message, localizer=self.localizer)
                )

    @override
    async def ask_confirmation(
        self,
        statement: ResolvableLocalizable,
        *,
        default: bool = False,
        stdin: TextIO | None = None,
    ) -> bool:
        self.assert_alive()
        return Confirm.ask(
            resolve_localized(statement, localizer=self.localizer),
            console=self._console,
            default=default,
            stream=stdin,
        )

    @overload
    async def ask_input(
        self,
        question: ResolvableLocalizable,
        *,
        default: str | VoidType = Void,
        stdin: TextIO | None = None,
    ) -> str:
        pass

    @overload
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T],
        default: str | VoidType = Void,
        stdin: TextIO | None = None,
    ) -> T:
        pass

    @override
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T] | None = None,
        default: str | VoidType = Void,
        stdin: TextIO | None = None,
    ) -> str | T:
        self.assert_alive()
        ask_kwargs = {}
        if default is not Void:
            ask_kwargs["default"] = default
        value = cast(
            str,
            Prompt.ask(
                resolve_localized(question, localizer=self.localizer),
                console=self._console,
                stream=stdin,
                **ask_kwargs,
            ),
        )
        if assertion is None:
            return value
        return assertion(value)



"""
Sequence types and implementations.
"""

from abc import abstractmethod
from collections.abc import Iterable, MutableSequence
from typing import overload, override

from betty.collection import MutableCollection


class MutableResolvedSequence[ValueT, ResolvableValueT](
    MutableSequence[ValueT], MutableCollection[ValueT]
):
    """
    A mutable sequence of resolved values.
    """

    @override
    @abstractmethod
    def insert(self, index: int, value: ValueT | ResolvableValueT) -> None:
        pass

    @overload
    def __setitem__(self, index: int, value: ValueT | ResolvableValueT) -> None:
        pass

    @overload
    def __setitem__(
        self, index: slice, value: Iterable[ValueT | ResolvableValueT]
    ) -> None:
        pass

    @abstractmethod
    def __setitem__(self, index, value):
        pass

    @override
    @abstractmethod
    def extend(self, values: Iterable[ValueT | ResolvableValueT]) -> None:
        pass



"""
The (de)serialization API.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final

from betty.definition.human_facing import HumanFacingDefinition
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localizable.markup import AnyEnumeration
from betty.locale.localizable.plain import Plain
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.media_type import MediaType
    from betty.portable import PortableData


class SerializationError(HumanFacingException):
    """
    Raised when an error occurs during (de)serialization.
    """


class Serializer(ABC, Plugin["SerializerDefinition"]):
    """
    A serializer.
    """

    @classmethod
    @abstractmethod
    def media_type(cls) -> MediaType:
        """
        The media type this serializer can serialize.
        """

    @abstractmethod
    def load(self, serialized: str, /) -> PortableData:
        """
        Deserialize data.

        :raise SerializationError: Raised when the dump could not be loaded.
        """

    @abstractmethod
    def dump(self, portable: PortableData, /) -> str:
        """
        Serialize data.
        """


@final
@PluginTypeDefinition(
    "serializer",
    label=_("Serializer"),
    label_plural=_("Serializers"),
    label_countable=ngettext("{count} serializer", "{count} serializers"),
    discovery=[EntryPointDiscovery("betty.serializer")],
)
class SerializerDefinition(HumanFacingDefinition, PluginDefinition[Serializer]):
    """
    .. plugin_type:: serializer.
    """


def serializer_for(
    available_serializers: Sequence[SerializerDefinition], extension: str, /
) -> SerializerDefinition:
    """
    Get the serializer for the given file extension.
    """
    for available_serializer in available_serializers:
        if extension in available_serializer.cls.media_type().extensions:
            return available_serializer
    raise SerializationError(
        _(
            'Unsupported file "{unsupported_type}". Supported types are: {available_types}.'
        ).format(
            unsupported_type=extension,
            available_types=AnyEnumeration(
                *[
                    Plain("{extension} ({available_type})").format(
                        extension=extension, available_type=available_serializer.label
                    )
                    for available_serializer in available_serializers
                    for extension in available_serializer.cls.media_type().extensions
                ]
            ),
        )
    )



"""
The localizable API allows objects to be localized at the point of use.
"""

from __future__ import annotations

import decimal
from abc import ABC, abstractmethod
from collections.abc import Mapping, MutableMapping
from typing import TYPE_CHECKING, override
from warnings import warn

from babel import Locale

from betty.locale import HasLocale, HasLocaleStr, ResolvableLocale
from betty.locale.localize import DEFAULT_LOCALIZER, Localizer

if TYPE_CHECKING:
    from ty_extensions import Intersection


class _Localizable[T](ABC):
    @abstractmethod
    def format(self, **format_kwargs: ResolvableLocalizable) -> T:
        """
        Apply string formatting to the eventual localized string.

        The arguments are identical to those of :py:meth:`str.format`.

        :return:
            A new localizable object.
        """


class Localizable(_Localizable["Localizable"]):
    """
    A localizable object.

    Objects of this type can convert themselves to localized strings at the point of use.
    """

    @abstractmethod
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        """
        Localize ``self`` to a human-readable string.
        """

    @override
    def format(self, **format_kwargs: ResolvableLocalizable) -> Localizable:
        return _FormattedLocalizable(self, format_kwargs)

    @override
    def __str__(self) -> str:
        localized = self.localize(DEFAULT_LOCALIZER)
        warn(
            f'{type(self)} ("{localized}") SHOULD NOT be cast to a string. Instead, call {type(self)}.localize() to ensure it is always formatted in the desired locale.',
            stacklevel=2,
        )
        return localized


type LocalizableCount = int | float | decimal.Decimal
"""
A count to localize strings for.

Based on :py:meth:`babel.plural.PluralRule.__call__`.
"""


class CountableLocalizable(_Localizable["CountableLocalizable"]):
    """
    An object that can be localized for a specific count (number of things).
    """

    @abstractmethod
    def count(self, count: LocalizableCount, /) -> Localizable:
        """
        Create a localizable for the given count (number of things).

        Implementations MUST automatically format the returned localizable with a ``{count}`` argument set to ``count``.
        """

    @override
    def format(self, **format_kwargs: ResolvableLocalizable) -> CountableLocalizable:
        return _FormattedCountableLocalizable(self, format_kwargs)


class _FormattedLocalizable(Localizable):
    def __init__(
        self,
        localizable: Localizable,
        format_kwargs: Mapping[str, ResolvableLocalizable],
        /,
    ):
        self._localizable = localizable
        self._format_kwargs = dict(format_kwargs)

    @override
    def format(self, **format_kwargs: ResolvableLocalizable) -> Localizable:
        self._format_kwargs.update(format_kwargs)
        return self

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        return HasLocaleStr(
            self._localizable.localize(localizer).format(
                **{
                    format_kwarg_key: format_kwarg.localize(localizer)
                    if isinstance(format_kwarg, Localizable)
                    else format_kwarg
                    for format_kwarg_key, format_kwarg in self._format_kwargs.items()
                },
            )
        )


class _FormattedCountableLocalizable(CountableLocalizable):
    def __init__(
        self,
        localizable: CountableLocalizable,
        format_kwargs: Mapping[str, ResolvableLocalizable],
        /,
    ):
        self._localizable = localizable
        self._format_kwargs = format_kwargs

    @override
    def count(self, count: LocalizableCount, /) -> Localizable:
        return _FormattedLocalizable(
            self._localizable.count(count),
            {**self._format_kwargs, "count": str(count)},
        )


type StaticTranslationsMapping = MutableMapping[Locale | None, str]
"""
Static translations for :py:class:`betty.locale.localizable.static.StaticTranslations`.

Values are a string, or a mapping of locales to translations.

See :py:func:`betty.locale.localizable.assertion.assert_static_translations`.
"""


type ShorthandStaticTranslations = (
    MutableMapping[ResolvableLocale | None, str] | str | StaticTranslationsMapping
)
"""
Static translations for :py:class:`betty.locale.localizable.static.StaticTranslations`.

Values are a string, or a mapping of locales or language tags to translations.

See :py:func:`betty.locale.localizable.assertion.assert_static_translations`.
"""


type CountableStaticTranslationsMapping = MutableMapping[
    Locale, MutableMapping[str, str]
]
"""
Countable static translations for :py:class:`betty.locale.localizable.CountableStaticTranslations`.

Values are mappings of locales to mappings of CLDR plural tags to translations.

See :py:func:`betty.locale.localizable.assertion.assert_countable_static_translations`.
"""


type ShorthandCountableStaticTranslations = MutableMapping[
    ResolvableLocale, MutableMapping[str, str]
]
"""
Static translations for :py:class:`betty.locale.localizable.static.StaticTranslations`.

Values are mappings of locales or language tags to mappings of CLDR plural tags to translations.

See :py:func:`betty.locale.localizable.assertion.assert_static_translations`.
"""


type ResolvableLocalizable = Localizable | ShorthandStaticTranslations
"""
A localizable, or a type that can be converted into a localizable with :py:func:`betty.locale.localizable.ensure_localizable`.
"""


type ResolvableCountableLocalizable = (
    CountableLocalizable | ShorthandCountableStaticTranslations
)
"""
A countable localizable, or a type that can be converted into a countable localizable with :py:func:`betty.locale.localizable.ensure_countable_localizable`.
"""


def resolve_localizable(localizable: ResolvableLocalizable) -> Localizable:
    """
    Ensure that a localizable-like value is or is made to be an actual localizable.
    """
    if isinstance(localizable, Localizable):
        return localizable
    if isinstance(localizable, str):
        from betty.locale.localizable.plain import Plain

        return Plain(localizable)
    from betty.locale.localizable.static import StaticTranslations

    return StaticTranslations(localizable)


def resolve_countable_localizable(
    localizable: ResolvableCountableLocalizable,
) -> CountableLocalizable:
    """
    Ensure that a countable-localizable-like value is or is made to be an actual countable localizable.
    """
    if isinstance(localizable, CountableLocalizable):
        return localizable
    from betty.locale.localizable.static import CountableStaticTranslations

    return CountableStaticTranslations(localizable)



"""
Manage translations of built-in translatable strings.
"""

from __future__ import annotations

import gettext
from abc import ABC, abstractmethod
from contextlib import redirect_stdout, suppress
from io import BytesIO, StringIO
from pathlib import Path
from typing import TYPE_CHECKING, final, override

import aiofiles
from aiofiles.os import makedirs
from aiofiles.ospath import exists
from polib import pofile

import betty
import betty.dirs
from betty.hashid import hashid_file_meta
from betty.locale import (
    DEFAULT_LOCALE,
    ResolvableLocale,
    from_language_tag,
    resolve_locale,
    to_language_tag,
)
from betty.locale.babel import run_babel
from betty.locale.error import LocaleError
from betty.locale.localizable.gettext import _
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterable, Mapping, MutableMapping

    from babel import Locale

    from betty.asset import AssetRepository
    from betty.cache.file import BinaryFileCache
    from betty.user import User


async def _new_translation(
    locale: Locale, assets_directory_path: Path, *, user: User
) -> None:
    po_file_path = (
        assets_directory_path / "locale" / to_language_tag(locale) / "betty.po"
    )
    with redirect_stdout(StringIO()):
        if await exists(po_file_path):
            await user.message_information(
                _("Translations for {locale} already exist at {po_file_path}.").format(
                    locale=to_language_tag(locale), po_file_path=str(po_file_path)
                )
            )
            return

        await run_babel(
            "",
            "init",
            "--no-wrap",
            "-i",
            str(assets_directory_path / "locale" / "betty.pot"),
            "-o",
            str(po_file_path),
            "-l",
            str(locale),
            "-D",
            "betty",
        )
        await user.message_information(
            _("Translations for {locale} initialized at {po_file_path}.").format(
                locale=to_language_tag(locale), po_file_path=str(po_file_path)
            )
        )


async def update_dev_translations(
    *,
    _output_assets_directory_path_override: Path | None = None,
) -> None:
    """
    Update the translations for Betty itself.
    """
    source_directory_path = betty.dirs.ROOT_DIRECTORY_PATH / "betty"
    test_directory_path = source_directory_path / "tests"
    await _update_translations(
        set(find_source_files(source_directory_path, test_directory_path)),
        betty.dirs.ASSETS_DIRECTORY_PATH,
        _output_assets_directory_path_override,
    )


async def _update_translations(
    source_file_paths: set[Path],
    assets_directory_path: Path,
    _output_assets_directory_path_override: Path | None = None,
) -> None:
    """
    Update all existing translations based on changes in translatable strings.
    """
    # During production, the input and output paths are identical. During testing,
    # _output_assets_directory_path provides an alternative output, so the changes
    # to the translations can be tested in isolation.
    output_assets_directory_path = (
        _output_assets_directory_path_override or assets_directory_path
    )

    pot_file_path = output_assets_directory_path / "locale" / "betty.pot"
    await makedirs(pot_file_path.parent, exist_ok=True)

    await run_babel(
        "",
        "extract",
        "--no-location",
        "--width",
        # Weblate uses 77 characters.
        "77",
        "--sort-output",
        "-F",
        "babel.ini",
        "-o",
        str(pot_file_path),
        "--project",
        "Betty",
        "--copyright-holder",
        "Bart Feenstra & contributors",
        *map(str, {*source_file_paths, *find_source_files(assets_directory_path)}),
    )
    for input_po_file_path in Path(assets_directory_path).glob("locale/*/betty.po"):
        output_po_file_path = (
            output_assets_directory_path
            / input_po_file_path.relative_to(assets_directory_path)
        ).resolve()
        await makedirs(output_po_file_path.parent, exist_ok=True)
        output_po_file_path.touch()

        locale = resolve_locale(output_po_file_path.parent.name)
        await run_babel(
            "",
            "update",
            "--domain",
            "betty",
            "--input-file",
            str(pot_file_path),
            "--ignore-obsolete",
            "--locale",
            str(locale),
            "--no-fuzzy-matching",
            "--output-file",
            str(output_po_file_path),
        )


def find_source_files(
    source_directory_path: Path, *exclude_directory_paths: Path
) -> Iterable[Path]:
    """
    Find source files in a directory.
    """
    resolved_exclude_directory_paths = {
        exclude_directory_path.expanduser().resolve()
        for exclude_directory_path in exclude_directory_paths
    }
    for source_file_path in source_directory_path.expanduser().resolve().rglob("*"):
        source_file_path = source_directory_path / source_file_path
        if set(resolved_exclude_directory_paths) & set(source_file_path.parents):
            continue
        if source_file_path.suffix in {".j2", ".py"}:
            yield source_file_path


class TranslationRepository(ABC):
    """
    Provide translations.
    """

    @property
    @abstractmethod
    def locales(self) -> Iterable[Locale]:
        """
        The available locales.
        """

    @abstractmethod
    def get(self, locale: ResolvableLocale) -> gettext.NullTranslations:
        """
        Get the translations for the given locale.
        """


@final
class StaticTranslationRepository(TranslationRepository):
    """
    Provide static translations.
    """

    def __init__(self, translations: Mapping[Locale, gettext.NullTranslations]):
        self._translations = translations

    @override
    @property
    def locales(self) -> Iterable[Locale]:
        return self._translations.keys()

    @override
    def get(self, locale: ResolvableLocale) -> gettext.NullTranslations:
        locale = resolve_locale(locale)
        try:
            return self._translations[locale]
        except KeyError:
            raise UntranslatedLocale(locale) from None


DEFAULT_TRANSLATION_REPOSITORY = StaticTranslationRepository(
    {DEFAULT_LOCALE: gettext.NullTranslations()}
)
"""
The translation repository for the default locale.
"""


@final
class ProxyTranslationRepository(TranslationRepository):
    """
    Provide translations from upstream repositories.
    """

    def __init__(self, *upstreams: TranslationRepository):
        self._upstreams = upstreams
        self._translations: MutableMapping[Locale, gettext.NullTranslations] = {}

    @override
    @property
    def locales(self) -> Iterable[Locale]:
        for upstream in self._upstreams:
            yield from upstream.locales

    @override
    def get(self, locale: ResolvableLocale) -> gettext.NullTranslations:
        locale = resolve_locale(locale)
        try:
            return self._translations[locale]
        except KeyError:
            translations: gettext.NullTranslations | None = None
            for upstream in self._upstreams:
                try:
                    upstream_translations = upstream.get(locale)
                except UntranslatedLocale:
                    pass
                else:
                    if translations is None:
                        translations = upstream_translations
                    else:
                        translations.add_fallback(upstream_translations)
            if translations is None:
                raise UntranslatedLocale(locale) from None
            self._translations[locale] = translations
            return translations


@final
@threadsafe
class AssetTranslationRepository(TranslationRepository):
    """
    Provide translations from assets.
    """

    def __init__(self, assets: AssetRepository, cache: BinaryFileCache):
        self._assets = assets
        self._cache = cache
        self._translations: MutableMapping[Locale, gettext.NullTranslations] = {}
        self._locales: set[Locale] = {DEFAULT_LOCALE}
        self._bootstrapped = False

    async def bootstrap(self) -> None:
        """
        Bootstrap the available translations.
        """
        assert not self._bootstrapped
        for assets_directory_path in reversed(self._assets.directories):
            for po_file_path in assets_directory_path.glob("locale/*/betty.po"):
                self._locales.add(from_language_tag(po_file_path.parent.name))
        for locale in self._locales:
            await self._build_translation(locale)
        self._bootstrapped = True

    @override
    @property
    def locales(self) -> Iterable[Locale]:
        assert self._bootstrapped
        return self._locales

    @override
    def get(self, locale: ResolvableLocale) -> gettext.NullTranslations:
        locale = resolve_locale(locale)
        try:
            return self._translations[locale]
        except KeyError:
            self._translations[locale] = gettext.NullTranslations()
            return self._translations[locale]

    async def _build_translation(self, locale: Locale) -> gettext.NullTranslations:
        translations = gettext.NullTranslations()
        for assets_directory_path in reversed(self._assets.directories):
            opened_translations = await self._open_translations(
                locale, assets_directory_path
            )
            if opened_translations:
                opened_translations.add_fallback(translations)
                translations = opened_translations
        self._translations[locale] = translations
        return self._translations[locale]

    async def _open_translations(
        self, locale: Locale, assets_directory_path: Path
    ) -> gettext.GNUTranslations | None:
        po_file_path = (
            assets_directory_path / "locale" / to_language_tag(locale) / "betty.po"
        )
        try:
            translation_version = await hashid_file_meta(po_file_path)
        except FileNotFoundError:
            return None
        cache_directory_path = self._cache.path / "locale" / translation_version
        mo_file_path = cache_directory_path / "betty.mo"

        with suppress(FileNotFoundError):
            async with aiofiles.open(mo_file_path, "rb") as f:
                return gettext.GNUTranslations(BytesIO(await f.read()))

        cache_directory_path.mkdir(exist_ok=True, parents=True)

        await run_babel(
            "",
            "compile",
            "-i",
            str(po_file_path),
            "-o",
            str(mo_file_path),
            "-l",
            str(resolve_locale(locale)),
            "-D",
            "betty",
        )
        async with aiofiles.open(mo_file_path, "rb") as f:
            return gettext.GNUTranslations(BytesIO(await f.read()))

    async def coverage(self, locale: ResolvableLocale) -> tuple[int, int]:
        """
        Get the translation coverage for the given locale.

        :return: A 2-tuple of the number of available translations and the
            number of translatable source strings.
        """
        translatables = {
            translatable async for translatable in self._get_translatables()
        }
        locale = resolve_locale(locale)
        if locale == DEFAULT_LOCALE:
            return len(translatables), len(translatables)
        translations = {
            translation async for translation in self._get_translations(locale)
        }
        return len(translations), len(translatables)

    async def _get_translatables(self) -> AsyncIterator[str]:
        for assets_directory_path in self._assets.directories:
            with suppress(FileNotFoundError):
                async with aiofiles.open(
                    assets_directory_path / "locale" / "betty.pot"
                ) as pot_data_f:
                    pot_data = await pot_data_f.read()
                    for entry in pofile(pot_data):
                        yield entry.msgid_with_context

    async def _get_translations(self, locale: Locale) -> AsyncIterator[str]:
        for assets_directory_path in reversed(self._assets.directories):
            with suppress(FileNotFoundError):
                async with aiofiles.open(
                    assets_directory_path
                    / "locale"
                    / to_language_tag(locale)
                    / "betty.po",
                    encoding="utf-8",
                ) as po_data_f:
                    po_data = await po_data_f.read()
                for entry in pofile(po_data):
                    if entry.translated():
                        yield entry.msgid_with_context


@final
class UntranslatedLocale(LocaleError):
    """
    Raised when no translations exist for a locale.
    """

    def __init__(self, locale: Locale, /):
        super().__init__(
            _("Untranslated locale {locale}.").format(locale=to_language_tag(locale))
        )



"""
Test utilities for :py:mod:`betty.serve`.
"""

from typing import Any, override

from betty.serve import ProjectServer, Server
from betty.test_utils.user import StaticUser


class NoOpServer(Server):
    """
    A server that does nothing.
    """

    def __init__(self, *_: Any, **__: Any):
        super().__init__(user=StaticUser())

    @override
    @property
    def public_url(self) -> str:
        return "https://example.com"

    @override
    async def start(self) -> None:
        pass

    @override
    async def stop(self) -> None:
        pass

    @override
    async def show(self) -> None:
        pass


class NoOpProjectServer(ProjectServer, NoOpServer):
    """
    A project server that does nothing.
    """



"""
Utilities for testing Jinja2 templates.
"""

from __future__ import annotations

from contextlib import AbstractAsyncContextManager, asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any

import aiofiles
from lxml.etree import ParserError
from lxml.html import document_fromstring

from betty.app import App
from betty.jinja import Environment
from betty.json.schema import AllOf, Ref
from betty.plugin import ResolvableId
from betty.project import Project
from betty.project.schema import ProjectSchema

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Callable, MutableMapping

    from jinja2 import Template

    from betty.extension import ExtensionDefinition


@asynccontextmanager
async def _assert_template(
    template_factory: Callable[[Environment, str], Template],
    template: str,
    *,
    data: MutableMapping[str, Any] | None = None,
    autoescape: bool | None = None,
    extensions: set[ResolvableId[ExtensionDefinition]] | None = None,
) -> AsyncIterator[tuple[str, Project]]:
    async with (
        App.new_isolated() as app,
        app,
        Project.new_isolated(app) as project,
    ):
        project.configuration.debug = True
        if extensions is not None:
            project.configuration.extensions.add(*extensions)
        async with project:
            if data is None:
                data = {}
            if "document" not in data:
                data["document"] = await project.new_document()
            jinja = await project.jinja
            if autoescape is not None:
                jinja.autoescape = autoescape
            rendered = await template_factory(jinja, template).render_async(**data)
            yield rendered, project


def assert_template_string(
    template: str,
    *,
    data: MutableMapping[str, Any] | None = None,
    autoescape: bool | None = None,
    extensions: set[ResolvableId[ExtensionDefinition]] | None = None,
) -> AbstractAsyncContextManager[tuple[str, Project]]:
    """
    Assert that a template string can be rendered.
    """
    return _assert_template(
        Environment.from_string,
        template,
        data=data,
        autoescape=autoescape,
        extensions=extensions,
    )


def assert_template_file(
    template: str,
    *,
    data: MutableMapping[str, Any] | None = None,
    autoescape: bool | None = None,
    extensions: set[ResolvableId[ExtensionDefinition]] | None = None,
) -> AbstractAsyncContextManager[tuple[str, Project]]:
    """
    Assert that a template file can be rendered.
    """
    return _assert_template(
        Environment.get_template,
        template,
        data=data,
        autoescape=autoescape,
        extensions=extensions,
    )


class _TemplateTestBase:
    extensions = set[ResolvableId]()
    """
    The extensions to enable before rendering the template.
    """


async def assert_betty_html(project: Project, url_path: str) -> Path:
    """
    Assert that an entity's HTML resource exists and is valid.
    """
    betty_html_file_path = project.www_directory / Path(url_path.lstrip("/"))
    async with aiofiles.open(betty_html_file_path) as f:
        betty_html = await f.read()
    try:
        document_fromstring(betty_html)
    except ParserError as e:
        raise ValueError(
            f'HTML parse error "{e}" in:\n{betty_html}'
        ) from None  # pragma: no cover

    return betty_html_file_path


async def assert_betty_json(project: Project, url_path: str, def_name: str) -> Path:
    """
    Assert that an entity's JSON resource exists and is valid.
    """
    import json

    betty_json_file_path = project.www_directory / Path(url_path.lstrip("/"))
    async with aiofiles.open(betty_json_file_path) as f:
        betty_json = await f.read()
    betty_json_data = json.loads(betty_json)

    project_schema = await ProjectSchema.new(project)
    # Somehow $ref cannot be top-level in our case, so wrap it.
    schema = AllOf(Ref(def_name))
    project_schema.embed(schema)

    schema.validate(betty_json_data)

    return betty_json_file_path



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Self, final, override

from betty.argparse import assertion_to_argument_type
from betty.assertion import assert_locale
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.console.project import add_project_argument
from betty.locale.localizable.gettext import _
from betty.locale.translation import project as translation_project
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from babel import Locale

    from betty.app import App
    from betty.project import Project


@final
@CommandDefinition("new-translation", label=_("Create a new translation"))
class NewTranslation(Manufacturable, Command):
    """
    .. plugin:: command:new-translation.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        localizer = await self._app.localizer
        command_function = await add_project_argument(
            parser, self._command_function, self._app
        )
        parser.add_argument(
            "locale",
            type=assertion_to_argument_type(assert_locale(), localizer=localizer),
        )
        return command_function

    async def _command_function(self, project: Project, locale: Locale) -> None:
        async with project:
            await translation_project.new_project_translation(
                locale, project, user=self._app.user
            )



"""
Access discovered plugins.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.json.schema import Enum
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.string import kebab_case_to_lower_camel_case

if TYPE_CHECKING:
    from collections.abc import Iterable


@final
class PluginIdSchema[
    PluginDefinitionT: PluginDefinition = PluginDefinition,
    PluginT: Plugin = Plugin,
](Enum):
    """
    The JSON schema for the IDs of the plugins in this repository.
    """

    def __init__(
        self,
        plugin_type: PluginTypeDefinition[PluginT, PluginDefinitionT],
        plugins: Iterable[PluginDefinition[PluginT]],
        /,
    ):
        label = plugin_type.label.localize(DEFAULT_LOCALIZER)
        super().__init__(
            *[plugin.id for plugin in plugins],  # noqa: A002
            def_name=kebab_case_to_lower_camel_case(plugin_type.id),
            title=label,
            description=f"A {label} plugin ID",
        )



"""
HTTP client rate limiting.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, final

from betty.concurrent import AsynchronizedLock, RateLimiter
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginTypeDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.plugin.ordered import OrderedPluginDefinition
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import MutableMapping, Sequence

    from aiohttp.client_middlewares import ClientHandlerType
    from aiohttp.client_reqrep import ClientRequest, ClientResponse


@final
@threadsafe
class RateLimitMiddleware:
    """
    HTTP client middleware to rate-limit requests.
    """

    def __init__(self, limits: Sequence[RateLimit], /):
        self._preferred_limits_and_limiters = [
            (limit, RateLimiter(*limit.limit)) for limit in limits
        ]
        self._default_limits_and_limiters: MutableMapping[
            tuple[str, str | None, int | None], RateLimiter
        ] = {}
        self._lock = AsynchronizedLock.new_threadsafe()

    async def __call__(
        self, request: ClientRequest, handler: ClientHandlerType
    ) -> ClientResponse:
        """
        Call the middleware.
        """
        request_limiter = self._get_matching_limiter(request)
        if not request_limiter:
            request_limiter = await self._get_default_limiter(request)
        await request_limiter.wait()
        return await handler(request)

    def _get_matching_limiter(self, request: ClientRequest) -> RateLimiter | None:
        for limit, limiter in self._preferred_limits_and_limiters:
            if limit.match(request):
                return limiter
        return None

    async def _get_default_limiter(self, request: ClientRequest) -> RateLimiter:
        default_key = (request.url.scheme, request.url.host, request.url.port)
        async with self._lock:
            try:
                return self._default_limits_and_limiters[default_key]
            except KeyError:
                request_limiter = RateLimiter(99, 1)
                self._default_limits_and_limiters[default_key] = request_limiter
                return request_limiter


class RateLimit(Plugin["RateLimitDefinition"]):
    """
    A rate limit for HTTP requests.
    """

    @abstractmethod
    def match(self, request: ClientRequest) -> bool:
        """
        Whether this limit matches the given request.
        """

    @property
    @abstractmethod
    def limit(self) -> tuple[int, int]:
        """
        The limit expressed as a 2-tuple of the maximum and the period (in seconds).
        """


@final
@PluginTypeDefinition(
    "http-rate-limit",
    label=_("HTTP client rate limit"),
    label_plural=_("HTTP client rate limits"),
    label_countable=ngettext(
        "{count} HTTP client rate limit", "{count} HTTP client rate limits"
    ),
    description=_(
        "Rate limits ensure that Betty's HTTP client does not make more requests to a web service than that service supports or allows, by enforcing a maximum number of requests per timeframe."
    ),
    discovery=[EntryPointDiscovery("betty.http_rate_limit")],
)
class RateLimitDefinition(OrderedPluginDefinition[RateLimit]):
    """
    .. plugin_type:: http-rate-limit.
    """



"""
Integrate Webpack with Jinja2.
"""

from collections.abc import MutableSet

from jinja2.runtime import Context

from betty.jinja import context_document


def _context_js_entry_points(context: Context) -> MutableSet[str]:
    document = context_document(context)
    try:
        js_entry_points = document["webpack_js_entry_points"]
        assert isinstance(js_entry_points, MutableSet)
        return js_entry_points  # ty:ignore[invalid-return-type]
    except KeyError:
        raise RuntimeError(
            "No `resource.webpack_js_entry_points` context variable exists in this Jinja2 template."
        ) from None



"""
Data types with human-readable description texts.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from betty.json.linked_data import (
    JsonLdObject,
    LinkedDataDumpableWithSchemaJsonLdObject,
    dump_context,
)
from betty.locale.localizable.gettext import _
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema
from betty.privacy import is_public
from betty.property import Optional

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable
    from betty.portable import PortableMapping
    from betty.project import Project


class HasDescription(LinkedDataDumpableWithSchemaJsonLdObject):
    """
    A resource with a description.
    """

    description = Optional(LocalizableProperty(label=_("Description")))
    """
    The description.
    """

    def __init__(
        self,
        *args: Any,
        description: ResolvableLocalizable | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.description = description

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "description",
            StaticTranslationsSchema(),
            False,
        )
        return schema

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        dump_context(portable, description="https://schema.org/description")
        if self.description is not None and is_public(self):
            portable["description"] = dump_linked_data(
                self.description, localizers=await project.public_localizers
            )
        return portable



"""
Provide functional programming utilities.
"""

from __future__ import annotations

import contextlib
from asyncio import sleep
from itertools import chain
from time import time
from typing import TYPE_CHECKING, Any, final

from betty.asyncio import resolve_await
from betty.typing import Void

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Iterable, Iterator


def map_suppress[T, U](
    raising_map: Callable[[T], U],
    exception_type: type[BaseException],
    items: Iterable[T],
    /,
) -> Iterator[U]:
    """
    Map values, skipping those for which the application of `raising_map` raises errors.
    """
    for item in items:
        try:
            yield raising_map(item)
        except exception_type:
            continue


class Do[**DoFP, DoFReturnT]:
    """
    A functional implementation of do-while functionality, with retries and timeouts.
    """

    def __init__(
        self,
        do: Callable[DoFP, DoFReturnT | Awaitable[DoFReturnT]],
        *do_args: DoFP.args,
        **do_kwargs: DoFP.kwargs,
    ):
        self._do = do
        self._do_args = do_args
        self._do_kwargs = do_kwargs

    async def until(
        self,
        *conditions: Callable[[DoFReturnT], None | bool | Awaitable[None | bool]],
        retries: int = 5,
        timeout: int = 300,
        interval: int | float = 0.1,
    ) -> DoFReturnT:
        """
        Perform the 'do' until it succeeds or as long as the given arguments allow.

        :param timeout: The timeout in seconds.
        :param interval: The interval between 'loops' in seconds.
        """
        start_time = time()
        while True:
            retries -= 1
            try:
                do_result = await resolve_await(
                    self._do(*self._do_args, **self._do_kwargs)
                )
                for condition in conditions:
                    if await resolve_await(condition(do_result)) is False:
                        raise RuntimeError(
                            f"Condition {condition} was not met for {do_result}."
                        )
            except Exception:
                if retries == 0:
                    raise
                if time() - start_time > timeout:
                    raise
                await sleep(interval)
            else:
                return do_result


def unique[ValueT](
    *values: Iterable[ValueT], key: Callable[[ValueT], Any] | None = None
) -> Iterator[ValueT]:
    """
    Yield the first occurrences of values in a sequence.

    For the purpose of filtering duplicate values from an iterable,
    this works similar to :py:class:`set`, except that this class
    supports non-hashable values. It is therefore slightly slower
    than :py:class:`set`.
    """
    seen_value_keys = []
    if key is None:
        key = passthrough
    for value in chain(*values):
        value_key = key(value)
        if value_key not in seen_value_keys:
            seen_value_keys.append(value_key)
            yield value


def passthrough[T](value: T, /) -> T:
    """
    Return the value.
    """
    return value


def suppress[**P, T](
    target: Callable[P, T], *exceptions: type[BaseException]
) -> Callable[P, T | Void]:
    """
    Return the value, but suppress any errors.
    """

    def _suppress(*target_args: P.args, **target_kwargs: P.kwargs) -> T | Void:
        with contextlib.suppress(*exceptions):
            return target(*target_args, **target_kwargs)
        return Void

    return _suppress


class ResultUnavailable(RuntimeError):
    """
    A :py:attr:`betty.functools.Result.result` is unavailable.
    """

    def __init__(self):
        super().__init__(
            "The result is unavailable because the target has not been called yet."
        )


@final
class Result[**P, T]:
    """
    Decorate a callable and store its return value or raised exception.
    """

    __slots__ = "_error", "_result", "_target"
    _error: BaseException
    _result: T

    def __init__(self, target: Callable[P, T], /):
        self._target = target

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        """
        Call the target.
        """
        try:
            self._result = self._target(*args, **kwargs)
            return self._result
        except BaseException as error:
            self._error = error
            raise

    def result(self) -> T:
        """
        Get the target's return value.

        If the target raised an exception, calling this method will re-raise the exception.
        """
        try:
            raise self._error
        except AttributeError:
            try:
                return self._result
            except AttributeError:
                raise ResultUnavailable from None



"""
Tools to build data types have citations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from betty.locale.localizable.gettext import _
from betty.model import Entity
from betty.model.association import BidirectionalToManySingleType, ToManyAssociates

if TYPE_CHECKING:
    from betty.ancestry.citation import Citation


class HasCitations(Entity):
    """
    An entity with citations that support it.
    """

    citations = BidirectionalToManySingleType["HasCitations", "Citation"](
        "betty.ancestry.citation:Citation",
        "facts",
        label=_("Citations"),
        description=_("The citations backing up the claims made by this entity"),
    )

    def __init__(
        self,
        *args: Any,
        citations: ToManyAssociates[Citation] | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if citations is not None:
            self.citations = citations



"""Provides Betty with SPDX data."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final, override

from betty.app import App
from betty.extension import Extension, ExtensionDefinition
from betty.license import LicenseDefinition
from betty.license.licenses import SpdxLicenseBuilder
from betty.locale.localizable.gettext import _
from betty.plugin.repository.static import StaticPluginRepository
from betty.service.factory import Manufacturable
from betty.service.provider import service
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    from betty.plugin.repository import PluginRepository


@final
@ExtensionDefinition(
    "spdx",
    label=_("SPDX licenses"),
    description=_(
        "Provide license plugins from the SPDX License List (https://spdx.org/licenses/) "
    ),
)
class Spdx(Manufacturable, Extension[App]):
    """
    .. plugin:: extension:spdx.
    """

    def __init__(self, *, app: App):
        super().__init__()
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app=app)

    @service
    async def license_repository(self) -> PluginRepository[LicenseDefinition]:
        """
        The SPDX licenses.
        """
        return StaticPluginRepository(
            LicenseDefinition,
            *[
                license
                async for license in SpdxLicenseBuilder(  # noqa: A001
                    binary_file_cache=self._app.binary_file_cache.with_scope("spdx"),
                    http_client=await self._app.http_client,
                    user=self._app.user,
                ).build()
            ],
        )



"""
Jobs for the Raspberry Mint extension.
"""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, override

import aiofiles

from betty.extension._theme.search import generate_search_index
from betty.extension.raspberry_mint import RaspberryMint
from betty.job import Job
from betty.locale import to_language_tag
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Chain
from betty.locale.localizable.plain import Plain
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.os import link_or_copy

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler
    from betty.project import Project


class _GenerateLogo(Job):
    def __init__(self, *, project: Project):
        super().__init__("raspberry-mint:generate-logo")
        self._project = project

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await link_or_copy(
            self._project.logo,
            self._project.www_directory / ("logo" + self._project.logo.suffix),
        )


class _GenerateSearchIndex(Job):
    _RESULT_CONTAINER_TEMPLATE = Plain("""
    <li class="d-flex gap-2 search-result">
        {{{ betty-search-result }}}
    </li>
    """)

    _RESULTS_CONTAINER_TEMPLATE = Chain(
        '<ul class="entity-list"><h3 class="h2">',
        _("Results ({{{ betty-search-results-count }}})"),
        "</h3>{{{ betty-search-results }}}</ul>",
    )

    def __init__(self, *, project: Project):
        super().__init__("raspberry-mint:generate-search-index")
        self._project = project

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await generate_search_index(
            self._project,
            self._RESULT_CONTAINER_TEMPLATE,
            self._RESULTS_CONTAINER_TEMPLATE,
            context=scheduler.context,
        )


class _GenerateWebmanifest(Job):
    def __init__(self, *, project: Project):
        super().__init__("raspberry-mint:generate-webmanifest")
        self._project = project

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        extensions = await self._project.extensions
        webmanifest = json.dumps(
            {
                "name": self._project.configuration.title.localize(DEFAULT_LOCALIZER),
                "icons": [
                    {"src": "/logo" + self._project.logo.suffix},
                ],
                "lang": to_language_tag(
                    self._project.configuration.default_locale.locale
                ),
                "theme_color": extensions[RaspberryMint].configuration.secondary_color,
                "background_color": "#ffffff",
                "display": "fullscreen",
            }
        )
        async with aiofiles.open(
            self._project.www_directory / "betty.webmanifest", "w"
        ) as f:
            await f.write(webmanifest)



"""
Object factories.
"""

from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Self, final, overload

from betty.asyncio import resolve_await
from betty.data import Data
from betty.exception import HumanFacingException
from betty.importlib import fully_qualified_name
from betty.locale.localizable.gettext import _
from betty.service.level import ServiceLevel
from betty.typing import Void, VoidType

if TYPE_CHECKING:
    from betty.portable import PortableData


class Manufacturable(ABC):
    """
    Allow this type to be instantiated using a :py:class:`betty.service.level.ServiceLevel`.
    """

    @classmethod
    @abstractmethod
    async def new(cls, services: ServiceLevel, /) -> Self:
        """
        Create a new instance using the given service level.
        """


type Manufacturer[T] = (
    Callable[[ServiceLevel], Awaitable[T] | T] | Callable[[], Awaitable[T] | T]
)


class DataManufacturable[DataT: Data](ABC):
    """
    A class that can be initialized using defined data.
    """

    @classmethod
    @abstractmethod
    async def new(cls, services: ServiceLevel, data: DataT, /) -> Self:
        """
        Create a new instance using the given service level and defined data.
        """

    @classmethod
    @abstractmethod
    def new_data_cls(cls) -> type[DataT]:
        """
        The object's defined data class.
        """


@final
class Factory:
    """
    The object factory.
    """

    def __init__(self, services: ServiceLevel, /):
        self._services = services

    @overload
    async def new[ManufacturableT](
        self,
        target: type[ManufacturableT],
        data: Data | PortableData | VoidType = Void,
        /,
    ) -> ManufacturableT:
        pass

    @overload
    async def new[T](
        self,
        target: Manufacturer[T],
        data: Data | PortableData | VoidType = Void,
        /,
    ) -> T:
        pass

    @overload
    async def new[T](
        self,
        target: type[T],
        data: Data | PortableData | VoidType = Void,
        /,
    ) -> T:
        pass

    async def new(self, target, data=Void):
        """
        Create a new instance.

        :raises FactoryError: raised when ``target`` could not be called.
        """
        if data is Void:
            if isinstance(target, type) and issubclass(target, Manufacturable):
                return await target.new(self._services)
            if callable(target):
                signature = inspect.signature(target)
                # If there is (at least) one positional argument, call the target with the service level.
                try:
                    signature.bind("")
                except TypeError:
                    result = target()
                else:
                    result = target(self._services)
                return await resolve_await(result)
            raise RuntimeError(f"Cannot create a new instance of {target}")
        if not isinstance(target, type) or not issubclass(target, DataManufacturable):
            raise HumanFacingException(
                _(
                    '"{target}" is not configurable, but configuration was given.'
                ).format(target=fully_qualified_name(target))
            )
        if not isinstance(data, Data):
            data = target.new_data_cls().data().porter.load(data)
        return await target.new(self._services, data)



"""
Expand an ancestry by deriving additional data from existing data.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final, override

from betty.extension import Extension, ExtensionDefinition
from betty.extension.deriver.jobs import DeriveAncestry
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.project.load import PostLoader
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler


@final
@ExtensionDefinition(
    "deriver",
    label="Deriver",
    description=_(
        "Create events such as births and deaths by deriving their details from existing information."
    ),
)
class Deriver(PostLoader, Manufacturable, Extension[Project]):
    """
    .. plugin:: extension:deriver.

    The ``deriver`` extension derives, or infers, events for people based on their existing events. For example, we know that someone's
    final disposition, such as a burial or cremation, comes after their death. If a person has a *burial* event without a
    date, and a *death* event with a date of *January 1, 1970*, the Deriver will update the *burial* event with the date
    range *sometime after January 1, 1970*.

    The Deriver works for every event type that declares it can be derived, and depending on which other event
    types it declares it comes before or after. This means that the behavior of this extension is complex, and dependent on
    the event types used within your site as well as the existing events for each person.
    """

    def __init__(self, *, project: Project):
        super().__init__()
        self._project = project

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(project=project)

    @override
    async def post_load(self, scheduler: Scheduler) -> None:
        await scheduler.add(DeriveAncestry(project=self._project))



"""
Static translations.
"""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, Self, final, override

from betty.assertion import assert_locale, assert_mapping, assert_or, assert_str
from betty.data.indicator.selector import Key
from betty.exception import reraise_with_indicator
from betty.locale import (
    HasLocale,
    HasLocaleStr,
    ResolvableLocale,
    negotiate_locale,
    plural_tags,
    resolve_locale,
    to_language_tag,
)
from betty.locale.localizable import (
    CountableLocalizable,
    CountableStaticTranslationsMapping,
    Localizable,
    LocalizableCount,
    ShorthandCountableStaticTranslations,
    ShorthandStaticTranslations,
    StaticTranslationsMapping,
)
from betty.locale.localizable.error import (
    InvalidPluralTag,
    MissingPluralPlaceholder,
    MissingPluralTag,
)
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import (
    AllEnumeration,
    Paragraphs,
    UnorderedList,
    do_you_mean,
)
from betty.portable import Portable, PortableData

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from babel import Locale
    from ty_extensions import Intersection

    from betty.locale.localize import Localizer


@final
class CountableStaticTranslations(CountableLocalizable, Portable):
    """
    A countable localizable backed by static translations.
    """

    _translations: CountableStaticTranslationsMapping

    def __init__(self, translations: ShorthandCountableStaticTranslations, /):
        from betty.assertion import assert_len

        assert_len(minimum=1)(translations)
        self._translations = {
            self._ensure_locale(locale, locale_translations): locale_translations
            for locale, locale_translations in translations.items()
        }

    @property
    def translations(self) -> CountableStaticTranslationsMapping:
        """
        The translations.
        """
        return dict(self._translations)

    def _ensure_locale(
        self, locale: ResolvableLocale, translations: Mapping[str, str]
    ) -> Locale:
        from betty.assertion import assert_len

        locale = resolve_locale(locale)
        with reraise_with_indicator(Key(to_language_tag(locale))):
            for plural_tag, translation in translations.items():
                with reraise_with_indicator(Key(plural_tag)):
                    assert_len(minimum=1)(translations)
                    if "{count}" not in translation:
                        raise MissingPluralPlaceholder(
                            Paragraphs(
                                _(
                                    "Missing `{{count}}` placeholder in {locale} plural translations"
                                ).format(locale=to_language_tag(locale)),
                                self._format_translations(translations),
                            )
                        )
            provided_plural_tags = set(translations.keys())
            locale_plural_tags = set(plural_tags(locale))
            invalid_plural_tags = provided_plural_tags - locale_plural_tags
            missing_plural_tags = locale_plural_tags - provided_plural_tags
            if invalid_plural_tags:
                raise InvalidPluralTag(
                    Paragraphs(
                        _(
                            "Invalid plural tag(s) {plural_tags} for {locale} translations."
                        ).format(
                            locale=to_language_tag(locale),
                            plural_tags=AllEnumeration(
                                *self._format_plural_tags(invalid_plural_tags)
                            ),
                        ),
                        do_you_mean(*self._format_plural_tags(locale_plural_tags)),
                        self._format_translations(translations),
                        self._format_plural_rules_link(locale),
                    )
                )
            if missing_plural_tags:
                raise MissingPluralTag(
                    Paragraphs(
                        _(
                            "Missing plural tag(s) {plural_tags} for {locale} translations."
                        ).format(
                            locale=to_language_tag(locale),
                            plural_tags=AllEnumeration(
                                *self._format_plural_tags(missing_plural_tags)
                            ),
                        ),
                        self._format_translations(translations),
                        self._format_plural_rules_link(locale),
                    )
                )
            return locale

    def _format_plural_tags(self, plural_tags: Iterable[str]) -> Iterable[str]:
        return [f'"{plural_tag}"' for plural_tag in sorted(plural_tags)]

    def _format_plural_rules_link(self, locale: Locale) -> Localizable:
        return _("Read more at {url}").format(
            url=f"https://www.unicode.org/cldr/charts/latest/supplemental/language_plural_rules.html#{locale}"
        )

    def _format_translations(self, translations: Mapping[str, str]) -> Localizable:
        return UnorderedList(
            *[
                f"{plural_tag}: {translation}"
                for plural_tag, translation in translations.items()
            ]
        )

    @override
    def count(self, count: LocalizableCount, /) -> Localizable:
        return StaticTranslations(
            {
                locale: self._translations[locale][locale.plural_form(count)]
                for locale in self._translations
            }
        ).format(count=str(count))

    @override
    @classmethod
    def load(cls, portable: PortableData, /) -> Self:
        return cls(
            assert_mapping(
                assert_mapping(
                    assert_str(),
                    assert_str(),
                ),
                assert_locale(),
            )(portable)
        )

    @override
    def dump(self) -> PortableData:
        return {
            to_language_tag(locale): translations
            for locale, translations in self.translations.items()
        }  # ty:ignore[invalid-return-type]


@final
class StaticTranslations(Localizable, Portable):
    """
    A localizable backed by static translations.

    All configuration options
    -------------------------

    This configuration is either a single translation as a string,
    or multiple translations as a key-value mapping.

    A single translation
    ^^^^^^^^^^^^^^^^^^^^

    .. list-table::
       :align: left
       :stub-columns: 1

       * -  Type
         -  string

    A single translation can be set that is to be used for all languages.

    Example configuration:

    .. tab-set::

       .. tab-item:: YAML

          .. code-block:: yaml

              "I am a single translation, used for all languages"

       .. tab-item:: JSON

          .. code-block:: json

              "I am a single translation, used for all languages"

    Multiple translations
    ^^^^^^^^^^^^^^^^^^^^^

    .. list-table::
       :align: left
       :stub-columns: 1

       * -  Type
         -  mapping
                Keys (string) are `IETF BCP 47 language tags <https://en.wikipedia.org/wiki/IETF_language_tag>`_,
                and values (string) are human-readable translations.

    Example configuration:

    .. tab-set::

       .. tab-item:: YAML

          .. code-block:: yaml

              en-US: "I'm the English translation"
              nl-NL: "Ik ben de Nederlandse vertaling"

       .. tab-item:: JSON

          .. code-block:: json

              {
                "en-US": "I'm the English translation",
                "nl-NL": "Ik ben de Nederlandse vertaling",
              }
    """

    _translations: StaticTranslationsMapping

    def __init__(self, translations: ShorthandStaticTranslations, /):
        """
        :param translations: Keys are locales, values are translations.
        """
        from betty.assertion import assert_len

        super().__init__()
        assert_len(minimum=1)(translations)
        self._translations = (
            {None: translations}
            if isinstance(translations, str)
            else {
                None if locale is None else resolve_locale(locale): translation
                for locale, translation in translations.items()
            }
        )
        assert len(self._translations) > 0

    @property
    def translations(self) -> StaticTranslationsMapping:
        """
        The translations.
        """
        return dict(self._translations)

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        if len(self._translations) > 1:
            available_locales = tuple(filter(None, self._translations.keys()))
            negotiated_locale = negotiate_locale(localizer.locale, available_locales)
            if negotiated_locale is not None:
                return HasLocaleStr(
                    self._translations[negotiated_locale], locale=negotiated_locale
                )
        locale, translation = next(iter(self._translations.items()))
        return HasLocaleStr(translation, locale=locale)

    @classmethod
    def resolve(cls, other: Localizable, localizers: Iterable[Localizer], /) -> Self:
        """
        Create a new instance from another :py:class`betty.locale.localizable.Localizable`.
        """
        if type(other) is cls:
            return other  # ty:ignore[invalid-return-type]
        return cls(
            {localizer.locale: other.localize(localizer) for localizer in localizers}
        )

    @override
    @classmethod
    def load(cls, portable: PortableData, /) -> Self:
        return cls(
            assert_or(
                assert_str().chain(lambda translation: {None: translation}),
                assert_mapping(assert_str(), assert_locale()),
            )(portable)
        )

    @override
    def dump(self) -> PortableData:
        if len(self.translations) == 1:
            with suppress(KeyError):
                # Explicitly cast to a string because pyyaml cannot dump ``str`` subclasses.
                return str(self.translations[None])
        return {
            # Explicitly cast to a string because pyyaml cannot dump ``str`` subclasses.
            to_language_tag(locale): str(translation)
            for locale, translation in self.translations.items()
        }



"""
Test utilities for :py:mod:`betty.ancestry.has_file_references`.
"""

from betty.ancestry.has_file_references import HasFileReferences
from betty.model import EntityDefinition
from betty.test_utils.locale.localizable import (
    DUMMY_COUNTABLE_LOCALIZABLE,
    DUMMY_LOCALIZABLE,
)


@EntityDefinition(
    "dummy-has-file-references",
    label=DUMMY_LOCALIZABLE,
    label_plural=DUMMY_LOCALIZABLE,
    label_countable=DUMMY_COUNTABLE_LOCALIZABLE,
)
class DummyHasFileReferences(HasFileReferences):
    """
    A dummy :py:class:`betty.ancestry.has_file_references.HasFileReferences` entity.
    """



"""
Test utilities for :py:mod:`betty.extension`.
"""

from typing import final

from betty.extension import Extension, ExtensionDefinition


class _DummyExtension(Extension):
    # Provide an initializer without arguments so the factory can call it.
    def __init__(self):
        super().__init__()


@final
@ExtensionDefinition("dummy-one", label="Dummy One")
class DummyExtensionOne(_DummyExtension):
    """
    A dummy :py:class:`betty.extension.Extension` implementation.
    """


@final
@ExtensionDefinition("dummy-two", label="Dummy Two")
class DummyExtensionTwo(_DummyExtension):
    """
    A dummy :py:class:`betty.extension.Extension` implementation.
    """



from __future__ import annotations  # noqa: D100

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

import betty.extension.demo as stddemo
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.extension.demo.project import create_project
from betty.job import Context
from betty.locale.localizable.gettext import _
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from betty.app import App


@final
@CommandDefinition("demo", label=_("Explore a demonstration site"))
class Demo(Manufacturable, Command):
    """
    .. plugin:: command:demo.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        parser.add_argument(
            "--path",
            help="The path to the project directory to generate the demonstration site into instead of serving the site in a browser window.",
        )
        parser.add_argument(
            "--url",
            help="The site's public project URL. Used only when `--path` is given.",
        )
        return self._command_function

    async def _command_function(self, *, path: str | None, url: str | None) -> None:
        from betty.extension.demo.serve import DemoServer

        if path is None:
            async with DemoServer(app=self._app) as server:
                await server.show()
                while True:
                    await asyncio.sleep(999)
        else:
            project = await create_project(self._app, Path(path))
            if url is not None:
                project.configuration.url = url
            async with (
                project,
                project.app.user.message_progress(_("Generating site...")) as progress,
            ):
                context = Context(progress=progress)
                await stddemo.generate_with_cleanup(project, context=context)



"""
Entity references.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.data import Data, Sample
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.str import StrDefinition
from betty.locale.localizable.gettext import _
from betty.machine_name import MachineNameProperty
from betty.plugin import resolve_id
from betty.property import Property

if TYPE_CHECKING:
    from betty.model import EntityDefinition
    from betty.plugin import ResolvableId


@final
@ObjectDefinition(
    label=_("Entity reference"),
    samples=[
        lambda: Sample(EntityReference("person", "123"), label="Default"),
    ],
)
class EntityReference(Data):
    """
    A reference to an entity of any type.

    .. data:: betty.model.reference:EntityReference
    """

    type = MachineNameProperty(label=_("Entity type"))
    """
    The type of the referenced entity. 
    """

    id = Property(StrDefinition(label=_("Entity ID")))
    """
    The ID of the referenced entity.
    """

    def __init__(
        self,
        /,
        type: ResolvableId[EntityDefinition],  # noqa: A002
        id: str,  # noqa: A002
    ):
        super().__init__()
        self.type = resolve_id(type)
        self.id = id



"""
Project support for the Console.
"""

import argparse
from contextlib import suppress
from pathlib import Path
from typing import Any

from betty.app import App
from betty.argparse import assertion_to_argument_type
from betty.assertion import assert_path
from betty.console.command import CommandFunction
from betty.error import FileNotFound
from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.portable.file import assert_load_file
from betty.project import Project
from betty.project.data import ProjectConfiguration
from betty.serde import SerializerDefinition
from betty.service.level import UNIVERSE
from betty.user import User


class ConfigurationFileNotFound(HumanFacingException):
    """
    Raised when no configuration file could be found.
    """


async def add_project_argument(
    parser: argparse.ArgumentParser,
    command_function: CommandFunction,
    app: App,
    *,
    required: bool = True,
) -> CommandFunction:
    """
    Add an argument to load a :py:class:`betty.project.Project` into a ``project`` keyword argument.
    """
    localizer = await app.localizer
    parser.add_argument(
        "-p",
        "--project",
        dest="project_configuration_file_path",
        help=localizer._(
            "The path to a Betty project directory or configuration file. Defaults to {default} in the current working directory."
        ).format(
            default=f"betty.{'|'.join(extension[1:] for serializer in await app.plugins.plugins(SerializerDefinition) for extension in serializer.cls.media_type().extensions)}"
        ),
        type=assertion_to_argument_type(assert_path(), localizer=localizer),
    )

    async def _command_function_with_project_argument(
        *, project_configuration_file_path: Path | None = None, **kwargs: Any
    ) -> None:
        project: Project | None
        try:
            (
                configuration,
                project_configuration_file_path,
            ) = await _read_project_configuration(
                project_configuration_file_path, app.user
            )
        except ConfigurationFileNotFound:
            if required:
                raise
            project = None
        else:
            project = Project(
                app, project_configuration_file_path, configuration=configuration
            )
        return await command_function(project=project, **kwargs)

    return _command_function_with_project_argument


async def _read_project_configuration(
    provided_configuration_file_path_str: Path | None, user: User
) -> tuple[ProjectConfiguration, Path]:
    project_directory_path = Path.cwd()
    if provided_configuration_file_path_str is None:
        try_configuration_file_paths = [
            project_directory_path / f"betty{extension}"
            for serializer in await UNIVERSE.plugins.plugins(SerializerDefinition)
            for extension in serializer.cls.media_type().extensions
        ]
        for try_configuration_file_path in try_configuration_file_paths:
            with suppress(FileNotFound):
                return await _read_project_configuration_file(
                    try_configuration_file_path, user
                )
        raise ConfigurationFileNotFound(
            _(
                "Could not find any of the following configuration files in {project_directory_path}: {configuration_file_names}."
            ).format(
                configuration_file_names=", ".join(
                    str(try_configuration_file_path.relative_to(project_directory_path))
                    for try_configuration_file_path in try_configuration_file_paths
                ),
                project_directory_path=str(project_directory_path),
            )
        )
    return await _read_project_configuration_file(
        (project_directory_path / provided_configuration_file_path_str)
        .expanduser()
        .resolve(),
        user,
    )


async def _read_project_configuration_file(
    configuration_file_path: Path, user: User
) -> tuple[ProjectConfiguration, Path]:
    assert_configuration = await assert_load_file()
    try:
        portable = assert_configuration(configuration_file_path)
    except HumanFacingException as error:
        await user.message_debug(error)
        raise
    else:
        await user.message_information_details(
            _("Loaded the configuration from {configuration_file_path}.").format(
                configuration_file_path=str(configuration_file_path)
            ),
        )
        return ProjectConfiguration.data().porter.load(
            portable
        ), configuration_file_path



"""
Tree content.
"""

from typing import Self, override

from betty.ancestry.person import Person
from betty.content import ContentDefinition
from betty.content.contents import Template, TemplateBuild
from betty.document import Document
from betty.extension.trees import Trees
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.service.factory import Manufacturable
from betty.service.requirement.extension import require_extension
from betty.service.requirement.project import require_project


@ContentDefinition("trees-tree", label=_("Family tree"))
class Tree(Template, Manufacturable):
    """
    An interactive family tree.

    .. plugin:: content:trees-tree
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(Trees, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, Person):
            return "component/trees/tree.html.j2", {
                "person": document.resource,
            }
        return None



"""
Localizable data.
"""

from __future__ import annotations

from typing import final

from betty.classtools import Singleton
from betty.data import DataDefinition
from betty.importlib import fully_qualified_name
from betty.locale.localizable import CountableLocalizable, Localizable
from betty.locale.localizable.gettext import _
from betty.locale.localizable.plain import Plain
from betty.locale.localizable.static import (
    CountableStaticTranslations,
    StaticTranslations,
)
from betty.portable import CallbackPorter, PortableData
from betty.portable.error import NotPortable


@final
class LocalizableDefinition(DataDefinition[Localizable], Singleton):
    """
    The data definition for :py:class:`betty.locale.localizable.Localizable`.
    """

    def __init__(self):
        super().__init__(
            cls=Localizable,
            label=_("A localizable string"),
            porter=CallbackPorter(StaticTranslations.load, self._dump),
        )

    def _dump(self, data: Localizable) -> PortableData:
        if isinstance(data, Plain):
            data = StaticTranslations({data.locale: data.text})
        if isinstance(data, StaticTranslations):
            return data.dump()
        raise NotPortable(
            Plain(
                "Only static translations and plain text can be dumped to portable data, not `{localizable}` objects."
            ).format(localizable=fully_qualified_name(type(data)))
        )


@final
class CountableLocalizableDefinition(DataDefinition[CountableLocalizable], Singleton):
    """
    The data definition for :py:class:`betty.locale.localizable.CountableLocalizable`.
    """

    def __init__(self):
        super().__init__(
            cls=CountableLocalizable,
            label=_("A countable localizable string"),
            porter=CallbackPorter(CountableStaticTranslations.load, self._dump),
        )

    def _dump(self, data: CountableLocalizable) -> PortableData:
        if isinstance(data, CountableStaticTranslations):
            return data.dump()
        raise NotPortable(
            Plain(
                "Only static translations and plain text can be dumped to portable data, not `{localizable}` objects."
            ).format(localizable=fully_qualified_name(type(data)))
        )



"""
Provide :py:class:`betty.license.License` plugins.
"""

import re
import tarfile
from asyncio import to_thread
from collections.abc import AsyncIterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from io import BytesIO
from json import loads
from pathlib import Path
from typing import final, override

import aiofiles
from aiohttp import ClientError, ClientSession

from betty.cache.file import BinaryFileCache
from betty.exception import HumanFacingException
from betty.license import License, LicenseDefinition
from betty.locale.localizable import Localizable
from betty.locale.localizable.gettext import _
from betty.locale.localizable.plain import Plain
from betty.machine_name import MachineName
from betty.user import User


@final
@LicenseDefinition("all-rights-reserved", label=_("All rights reserved"))
class AllRightsReserved(License):
    """
    .. plugin:: license:all-rights-reserved.
    """

    @property
    @override
    def summary(self) -> Localizable:
        return self.plugin().label

    @property
    @override
    def text(self) -> Localizable:
        return _(
            "No part may be reproduced or distributed in any form or by any means, without express written permission from the copyright holder, or unless permitted by copyright law."
        )


@final
@LicenseDefinition("public-domain", label=_("Public domain"))
class PublicDomain(License):
    """
    .. plugin:: license:public-domain.
    """

    @property
    @override
    def summary(self) -> Localizable:
        return self.plugin().label

    @property
    @override
    def text(self) -> Localizable:
        return _(
            "Works in the public domain can be used or referenced without permission, because nobody holds any exclusive rights over these works (anymore)."
        )


_SPDX_LICENSE_ID_PATTERN = re.compile(r"[^a-z0-9-]")


def spdx_license_id_to_license_id(spdx_license_id: str) -> MachineName:
    """
    Get the Betty license plugin ID for the given SPDX license ID.
    """
    return MachineName(
        f"spdx-{_SPDX_LICENSE_ID_PATTERN.sub('--', spdx_license_id.lower())}"
    )


class SpdxLicenseBuilder:
    """
    Build licenses from the `SPDX License List <https://spdx.org/licenses/>`_.
    """

    VERSION = "3.27.0"
    URL = (
        f"https://github.com/spdx/license-list-data/archive/refs/tags/v{VERSION}.tar.gz"
    )

    def __init__(
        self,
        *,
        http_client: ClientSession,
        user: User,
        binary_file_cache: BinaryFileCache,
    ):
        self._http_client = http_client
        self._user = user
        self._cache_directory_path = (
            binary_file_cache.with_scope("spdx-licenses").with_scope(self.VERSION).path
        )

    async def build(self) -> AsyncIterable[LicenseDefinition]:
        """
        Build the licenses.
        """
        if not self._cache_directory_path.exists():
            try:
                spdx_licenses_response = await self._http_client.get(self.URL)
                spdx_licenses_data_tar = await spdx_licenses_response.read()
            except ClientError:
                await self._user.message_warning(
                    _("Betty could not load the SPDX licenses")
                )
                return

            await to_thread(
                self._extract_licenses,
                spdx_licenses_data_tar,
                self._cache_directory_path,
            )

        async with aiofiles.open(
            self._cache_directory_path
            / f"license-list-data-{self.VERSION}"
            / "json"
            / "licenses.json",
            encoding="utf-8",
        ) as spdx_licenses_data_f:
            spdx_licenses_data_json = await spdx_licenses_data_f.read()
        spdx_data = loads(spdx_licenses_data_json)
        assert isinstance(spdx_data, Mapping)

        spdx_licenses_data = spdx_data["licenses"]
        assert isinstance(spdx_licenses_data, Sequence)

        for spdx_license_data in spdx_licenses_data:
            assert isinstance(spdx_license_data, Mapping)

            if spdx_license_data.get("isDeprecatedLicenseId", False):
                continue

            spdx_license_id = spdx_license_data["licenseId"]
            assert isinstance(spdx_license_id, str)
            spdx_license_id_to_license_id(spdx_license_id)

            spdx_reference = spdx_license_data["reference"]
            assert isinstance(spdx_reference, str)

            yield await self._build_license(spdx_license_id, spdx_reference)

    async def _build_license(self, license_id: str, url: str) -> LicenseDefinition:
        async with aiofiles.open(
            self._cache_directory_path
            / f"license-list-data-{self.VERSION}"
            / "json"
            / "details"
            / f"{license_id}.json",
            encoding="utf-8",
        ) as spdx_license_data_f:
            spdx_license_data_json = await spdx_license_data_f.read()

        with self._catch_json_errors():
            spdx_license_data = loads(spdx_license_data_json)
            assert isinstance(spdx_license_data, Mapping)

            license_name = spdx_license_data["name"]
            assert isinstance(license_name, str)

            license_text = spdx_license_data["licenseText"]
            assert isinstance(license_text, str)

            @LicenseDefinition(
                spdx_license_id_to_license_id(license_id), label=license_name
            )
            class _SpdxLicense(License):
                @override
                @property
                def summary(self) -> Localizable:
                    return self.plugin().label

                @override
                @property
                def text(self) -> Localizable:
                    return Plain(license_text)

                @override
                @property
                def url(self) -> Localizable | None:
                    return Plain(url)

            return _SpdxLicense.plugin()

    @classmethod
    def _extract_licenses(
        cls, spdx_licenses_data_tar: bytes, cache_directory_path: Path
    ):
        with tarfile.open(
            fileobj=BytesIO(spdx_licenses_data_tar), mode="r:gz"
        ) as tar_file:
            tar_file.extractall(
                cache_directory_path,
                members=[
                    tar_file.getmember(
                        f"license-list-data-{cls.VERSION}/json/licenses.json"
                    ),
                    *[
                        tar_info
                        for tar_info in tar_file.getmembers()
                        if tar_info.name.startswith(
                            f"license-list-data-{cls.VERSION}/json/details/"
                        )
                    ],
                ],
                filter="data",
            )

    @contextmanager
    def _catch_json_errors(self) -> Iterator[None]:
        try:
            yield
        except (AssertionError, LookupError) as error:
            raise HumanFacingException(
                Plain(f"Invalid JSON response received from {self.URL}")
            ) from error



"""
Describe, access, and manipulate arbitrary data.
"""

from __future__ import annotations

from functools import update_wrapper
from typing import TYPE_CHECKING, Any, Self, final, override

from betty.definition.cls import ClsDefinition
from betty.definition.human_facing import HumanFacingDefinition
from betty.importlib import fully_qualified_name
from betty.portable import (
    OptionalPorter,
    Portable,
    PortableData,
    PortablePorter,
    Porter,
)
from betty.portable.error import NotPortable
from betty.sample import Samplable, Sample, Samples, Size

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from ty_extensions import Intersection

    from betty.locale.localizable import ResolvableLocalizable


class DataDefinition[DataClsT, PortableDataT: PortableData = PortableData](
    HumanFacingDefinition, ClsDefinition[DataClsT]
):
    """
    A data definition.
    """

    def __init__(
        self,
        /,
        cls: type[DataClsT] | None = None,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        porter: Porter[DataClsT, PortableDataT] | None = None,
        samples: Iterable[
            Callable[[], Sample[DataClsT]]
            | Samples[DataClsT]
            | type[Intersection[DataClsT, Samplable]]
        ]
        | None = None,
    ):
        super().__init__(cls=cls, label=label, description=description)
        self._porter = porter
        self._samples = samples

    @property
    def porter(self) -> Porter[DataClsT, PortableDataT]:
        """
        The porter for the data.
        """
        if self._porter is None:
            if not issubclass(self.cls, Portable):
                raise NotPortable(
                    f"This definition does not have a porter. Either make the data class {fully_qualified_name(self.cls)} subclass {fully_qualified_name(Portable)}, or provide a porter when initializing the definition."
                )
            self._porter = PortablePorter(self.cls)
        return self._porter  # ty:ignore[invalid-return-type]

    @override
    def _set_cls(self, cls: type[DataClsT]) -> None:
        super()._set_cls(cls)
        if issubclass(cls, Data):
            cls.data = staticmethod(update_wrapper(lambda: self, cls.data))  # ty:ignore[invalid-assignment]

    @property
    def samples(self) -> Samples:
        """
        Any samples for this data.
        """
        if self._samples is None:
            if issubclass(self.cls, Samplable):
                return Samples([self.cls])
            return Samples(())
        return Samples(self._samples)


class Data[DataDefinitionT: DataDefinition = DataDefinition]:
    """
    A class that defines data for its instances.
    """

    @classmethod
    def data(cls) -> Intersection[DataDefinitionT, DataDefinition[Self]]:
        """
        Define the data for instances of this class.
        """
        raise NotImplementedError(
            f"{fully_qualified_name(cls)} was not decorated with {fully_qualified_name(DataDefinition)} or any subclass."
        )

    def __eq__(self, other: Any) -> bool:
        if type(self) is not type(other):
            return NotImplemented
        porter = type(self).data().porter
        return porter.dump(self) == porter.dump(other)


@final
class OptionalDefinition[DataClsT](DataDefinition[DataClsT | None]):
    """
    Wrap another data definition to make it optional, e.g. allow ``None``.
    """

    def __init__(self, wrapped: DataDefinition[DataClsT], /):
        super().__init__(
            cls=wrapped.cls,
            label=wrapped.label,
            description=wrapped.description,
            porter=OptionalPorter(wrapped.porter),  # ty:ignore[invalid-argument-type]
            samples=[
                lambda: Sample(None, label="Minimal", size=Size.MINIMAL),
                wrapped.samples,
            ],
        )
        self._wrapped = wrapped

    @property
    def wrapped(self) -> DataDefinition[DataClsT]:
        """
        The wrapped, required (non-optional) data definition.
        """
        return self._wrapped



"""
Data types to describe the relationships between places.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.ancestry.date import HasDate
from betty.ancestry.has_citations import HasCitations
from betty.locale.localizable.gettext import _, ngettext
from betty.model import Entity, EntityDefinition
from betty.model.association import BidirectionalToOne, ToOneAssociate

if TYPE_CHECKING:
    from betty.ancestry.place import Place


@final
@EntityDefinition(
    "enclosure",
    label=_("Enclosure"),
    label_plural=_("Enclosures"),
    label_countable=ngettext("{count} enclosure", "{count} enclosures"),
    public_facing=False,
)
class Enclosure(HasDate, HasCitations, Entity):
    """
    .. plugin:: entity:enclosure.
    """

    encloser = BidirectionalToOne["Enclosure", "Place"](
        "betty.ancestry.place:Place",
        "enclosees",
        label=_("Encloser"),
        description=_("The place that encloses or contains the enclosee"),
    )
    """
    The outer place.
    """

    enclosee = BidirectionalToOne["Enclosure", "Place"](
        "betty.ancestry.place:Place",
        "enclosers",
        label=_("Enclosee"),
        description=_("The place that is enclosed or contained by the encloser"),
    )
    """
    The inner place.
    """

    def __init__(
        self, enclosee: ToOneAssociate[Place], encloser: ToOneAssociate[Place]
    ):
        super().__init__()
        self.enclosee = enclosee
        self.encloser = encloser



"""
Adapters between Betty and Python sequences.
"""

from collections.abc import Callable, Iterable, MutableSequence, Sequence
from contextlib import suppress
from typing import Any, final, overload, override

from betty.collection.sequence import MutableResolvedSequence


class _ResolvedSequenceAdapter[ValueT, ResolvableValueT](Sequence[ValueT]):
    def __init__(
        self,
        upstream: Sequence[ValueT],
        *,
        value_resolver: Callable[[ValueT | ResolvableValueT], ValueT],
    ):
        self._upstream = upstream
        self._value_resolver = value_resolver

    @overload
    def __getitem__(self, index: int) -> ValueT:
        pass

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[ValueT]:
        pass

    @final
    def __getitem__(self, index):
        return self._upstream[index]

    @final
    def __len__(self):
        return len(self._upstream)

    @final
    def __contains__(self, value: Any) -> bool:
        with suppress(Exception):
            value = self._value_resolver(value)
        return value in self._upstream

    @final
    @override
    def index(self, value: Any, start: int = 0, stop: int = -1) -> int:
        with suppress(Exception):
            value = self._value_resolver(value)
        return self._upstream.index(value, start, stop)


@final
class ResolvedSequenceAdapter[ValueT, ResolvableValueT](
    _ResolvedSequenceAdapter[ValueT, ResolvableValueT]
):
    """
    Decorate another sequence to resolve any values before proxying them.
    """


@final
class MutableResolvedSequenceAdapter[ValueT, ResolvableValueT](
    _ResolvedSequenceAdapter[ValueT, ResolvableValueT],
    MutableResolvedSequence[ValueT, ResolvableValueT],
):
    """
    Decorate another sequence to resolve any values before proxying them.
    """

    _upstream: MutableSequence[ValueT]

    def __init__(
        self,
        upstream: MutableSequence[ValueT],
        *,
        value_resolver: Callable[[ValueT | ResolvableValueT], ValueT],
    ):
        super().__init__(upstream, value_resolver=value_resolver)

    @override
    def insert(self, index: int, value: ValueT | ResolvableValueT) -> None:
        self._upstream.insert(index, self._value_resolver(value))

    @overload
    def __setitem__(self, index: int, value: ValueT | ResolvableValueT) -> None:
        pass

    @overload
    def __setitem__(
        self, index: slice, value: Iterable[ValueT | ResolvableValueT]
    ) -> None:
        pass

    def __setitem__(self, index, value):
        if isinstance(index, int):
            self._upstream[index] = self._value_resolver(value)
        else:
            self._upstream[index] = map(self._value_resolver, value)

    def __delitem__(self, index: int | slice) -> None:
        del self._upstream[index]

    @override
    def extend(self, values: Iterable[ValueT | ResolvableValueT]) -> None:
        self._upstream.extend(map(self._value_resolver, values))



"""
Copyright notice configuration.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, final, override

from betty.copyright_notice import CopyrightNotice, CopyrightNoticeDefinition
from betty.data.aggregate.record.object import ObjectDefinition
from betty.locale.localizable.gettext import _
from betty.locale.localizable.property import LocalizableProperty
from betty.plugin.data import HumanFacingPluginDefinitionConfiguration
from betty.sample import Sample

if TYPE_CHECKING:
    from betty.locale.localizable import Localizable, ResolvableLocalizable


@final
@ObjectDefinition(
    label=_("Copyright notice configuration"),
    samples=[
        lambda: Sample(
            CopyrightNoticeDefinitionConfiguration(
                id="my-first-copyright-notice",
                label="My First Copyright Notice",
                summary="My First Copyright Notice is my first copyright notice",
                text="My First Copyright Notice is my first copyright notice, all rights are reserved.",
            ),
            label="Default",
        )
    ],
)
class CopyrightNoticeDefinitionConfiguration(
    HumanFacingPluginDefinitionConfiguration[CopyrightNoticeDefinition]
):
    """
    Configure a :py:class:`betty.copyright_notice.CopyrightNoticeDefinition`.

    .. data:: betty.project.data:CopyrightNoticeDefinitionConfiguration
    """

    summary = LocalizableProperty(label=_("Summary"))
    text = LocalizableProperty(label=_("Text"))

    def __init__(
        self,
        *,
        summary: ResolvableLocalizable,
        text: ResolvableLocalizable,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.summary = summary
        self.text = text

    @override
    def new_plugin(self) -> CopyrightNoticeDefinition:
        configuration = self

        @CopyrightNoticeDefinition(
            self.id,
            label=self.label,
            description=self.description,
        )
        class _ProjectConfigurationCopyrightNotice(CopyrightNotice):
            @override
            @property
            def summary(self) -> Localizable:
                return configuration.summary

            @override
            @property
            def text(self) -> Localizable:
                return configuration.text

        return _ProjectConfigurationCopyrightNotice.plugin()



"""
Gettext translations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast, overload, override

from betty.locale.localizable import CountableLocalizable, Localizable, LocalizableCount

if TYPE_CHECKING:
    from ty_extensions import Intersection

    from betty.locale import HasLocale
    from betty.locale.localize import Localizer


class _GettextLocalizable(Localizable):
    def __init__(
        self,
        gettext_method_name: str,
        *gettext_args: Any,
    ) -> None:
        self._gettext_method_name = gettext_method_name
        self._gettext_args = gettext_args

    @override
    def localize(self, localizer: Localizer, /) -> Intersection[HasLocale, str]:
        return cast(
            "Intersection[HasLocale, str]",
            getattr(localizer, self._gettext_method_name)(*self._gettext_args),
        )


class _CountableGettextLocalizable(CountableLocalizable):
    def __init__(
        self,
        gettext_method_name: str,
        *gettext_args: Any,
    ) -> None:
        self._gettext_method_name = gettext_method_name
        self._gettext_args = gettext_args

    @override
    def count(self, count: LocalizableCount, /) -> Localizable:
        return _GettextLocalizable(
            self._gettext_method_name, *self._gettext_args, count
        ).format(count=str(count))


def gettext(message: str, /) -> Localizable:
    """
    Like :py:meth:`gettext.gettext`.

    Positional arguments are identical to those of :py:meth:`gettext.gettext`.
    Keyword arguments are identical to those of :py:meth:`str.format`, except that
    any :py:class:`betty.locale.localizable.Localizable` will be localized before string
    formatting.
    """
    return _GettextLocalizable("gettext", message)


def _(message: str, /) -> Localizable:
    """
    Like :py:meth:`betty.locale.localizable.gettext`.

    Positional arguments are identical to those of :py:meth:`gettext.gettext`.
    Keyword arguments are identical to those of :py:meth:`str.format`, except that
    any :py:class:`betty.locale.localizable.Localizable` will be localized before string
    formatting.
    """
    return gettext(message)


@overload
def ngettext(message_singular: str, message_plural: str, n: int, /) -> Localizable:
    pass


@overload
def ngettext(
    message_singular: str, message_plural: str, n: None = None, /
) -> CountableLocalizable:
    pass


def ngettext(
    message_singular: str, message_plural: str, n: int | None = None, /
) -> Localizable | CountableLocalizable:
    """
    Like :py:meth:`gettext.ngettext`.

    Positional arguments are identical to those of :py:meth:`gettext.ngettext`.
    Keyword arguments are identical to those of :py:meth:`str.format`, except that
    any :py:class:`betty.locale.localizable.Localizable` will be localized before string
    formatting.

    Messages MUST have a ``{count}`` placeholder.
    """
    if n is None:
        return _CountableGettextLocalizable(
            "ngettext", message_singular, message_plural
        )
    return _GettextLocalizable("ngettext", message_singular, message_plural, n).format(
        count=str(n)
    )


def pgettext(context: str, message: str, /) -> Localizable:
    """
    Like :py:meth:`gettext.pgettext`.

    Positional arguments are identical to those of :py:meth:`gettext.pgettext`.
    Keyword arguments are identical to those of :py:meth:`str.format`, except that
    any :py:class:`betty.locale.localizable.Localizable` will be localized before string
    formatting.
    """
    return _GettextLocalizable("pgettext", context, message)


@overload
def npgettext(
    context: str, message_singular: str, message_plural: str, n: int, /
) -> Localizable:
    pass


@overload
def npgettext(
    context: str, message_singular: str, message_plural: str, n: None = None, /
) -> CountableLocalizable:
    pass


def npgettext(
    context: str, message_singular: str, message_plural: str, n: int | None = None, /
) -> Localizable | CountableLocalizable:
    """
    Like :py:meth:`gettext.npgettext`.

    Positional arguments are identical to those of :py:meth:`gettext.npgettext`.
    Keyword arguments are identical to those of :py:meth:`str.format`, except that
    any :py:class:`betty.locale.localizable.Localizable` will be localized before string
    formatting.
    """
    if n is None:
        return _CountableGettextLocalizable(
            "npgettext", context, message_singular, message_plural
        )
    return _GettextLocalizable(
        "npgettext", context, message_singular, message_plural, n
    ).format(count=str(n))
